# AUTO-GENERATED FILE. Do not edit. Run chalkpy stubgen to generate.
# fmt: off
# isort: skip_file
from __future__ import annotations

from chalk.features import DataFrame as DataFrame
from chalk.features import Features as Features
from chalk.features import Tags as Tags
from chalk.features.feature_set import FeaturesMeta as FeaturesMeta
from chalk.streams._windows import Windowed as __stubgen_chalk_streams___windows_Windowed
from chalk.utils.duration import Duration as Duration
from datetime import date as __stubgen_datetime_date
from datetime import datetime as __stubgen_datetime_datetime
from tests.client.test_client_serialization import Color as __stubgen_tests_client_test__client__serialization_Color
from tests.client.test_expand_features import Ankle as __stubgen_tests_client_test__expand__features_Ankle
from tests.client.test_expand_features import Foot as __stubgen_tests_client_test__expand__features_Foot
from tests.features.test_chained_feature_time import Homeowner as __stubgen_tests_features_test__chained__feature__time_Homeowner
from tests.features.test_chained_has_one import ExampleFraudOrg as __stubgen_tests_features_test__chained__has__one_ExampleFraudOrg
from tests.features.test_chained_has_one import ExampleFraudUser as __stubgen_tests_features_test__chained__has__one_ExampleFraudUser
from tests.features.test_df import Foo as __stubgen_tests_features_test__df_Foo
from tests.features.test_df import Topping as __stubgen_tests_features_test__df_Topping
from tests.features.test_df import ToppingPrice as __stubgen_tests_features_test__df_ToppingPrice
from tests.features.test_features import SingleChildFS as __stubgen_tests_features_test__features_SingleChildFS
from tests.features.test_features import SingleParentFS as __stubgen_tests_features_test__features_SingleParentFS
from tests.features.test_iter import NoFunFeatures as __stubgen_tests_features_test__iter_NoFunFeatures
from tests.serialization.test_codec import Color as __stubgen_tests_serialization_test__codec_Color
from tests.serialization.test_codec import CustomClass as __stubgen_tests_serialization_test__codec_CustomClass
from tests.serialization.test_codec import MyDataclass as __stubgen_tests_serialization_test__codec_MyDataclass
from typing import Optional as Optional
from typing import Protocol as Protocol
from typing import Type as Type
from typing import Union as Union
from typing import overload as overload

class WowFSMetaclass(FeaturesMeta):
    @property
    def something(self) -> Type[str]: ...

    @property
    def something_else(self) -> Type[str]: ...

    @property
    def nocomment(self) -> Type[str]: ...

    @property
    def nope(self) -> Type[str]: ...

    @property
    def assigned(self) -> Type[str]: ...

    @property
    def bizarre(self) -> Type[str]: ...

    @property
    def goofy(self) -> Type[str]: ...

    @property
    def assigned_comment(self) -> Type[str]: ...

    @property
    def explicit(self) -> Type[str]: ...

    @property
    def assigned_comment_multiline(self) -> Type[str]: ...

    @property
    def time(self) -> Type[__stubgen_datetime_datetime]: ...

    @property
    def nope_nope(self) -> Type[__stubgen_datetime_datetime]: ...

class WowFS(Features, metaclass=WowFSMetaclass):
    def __init__(
        self,
        something: str = ...,
        something_else: str = ...,
        nocomment: str = ...,
        nope: str = ...,
        assigned: str = ...,
        bizarre: str = ...,
        goofy: str = ...,
        assigned_comment: str = ...,
        explicit: str = ...,
        assigned_comment_multiline: str = ...,
        time: __stubgen_datetime_datetime = ...,
        nope_nope: __stubgen_datetime_datetime = ...,
    ):
        self.something: str
        self.something_else: str
        self.nocomment: str
        self.nope: str
        self.assigned: str
        self.bizarre: str
        self.goofy: str
        self.assigned_comment: str
        self.explicit: str
        self.assigned_comment_multiline: str
        self.time: __stubgen_datetime_datetime
        self.nope_nope: __stubgen_datetime_datetime

class WowFSProtocol(Protocol):
    something: str
    something_else: str
    nocomment: str
    nope: str
    assigned: str
    bizarre: str
    goofy: str
    assigned_comment: str
    explicit: str
    assigned_comment_multiline: str
    time: __stubgen_datetime_datetime
    nope_nope: __stubgen_datetime_datetime

class TacoMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def user_id(self) -> Type[str]: ...

    @property
    def price(self) -> Type[int]: ...

    @property
    def maybe_price(self) -> Type[Union[int, None]]: ...

    @property
    def hat(self) -> Type[str]: ...

    @property
    def topping_id(self) -> Type[str]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

    @property
    def nicknames(self) -> Type[DataFrame]: ...

    @property
    def topping(self) -> Type[__stubgen_tests_features_test__df_Topping]: ...

    @property
    def foo(self) -> Type[__stubgen_tests_features_test__df_Foo]: ...

class Taco(Features, metaclass=TacoMetaclass):
    def __init__(
        self,
        id: str = ...,
        user_id: str = ...,
        price: int = ...,
        maybe_price: Union[int, None] = ...,
        hat: str = ...,
        topping_id: str = ...,
        ts: __stubgen_datetime_datetime = ...,
        nicknames: DataFrame = ...,
        topping: __stubgen_tests_features_test__df_Topping = ...,
        foo: __stubgen_tests_features_test__df_Foo = ...,
    ):
        self.id: str
        self.user_id: str
        self.price: int
        self.maybe_price: Union[int, None]
        self.hat: str
        self.topping_id: str
        self.ts: __stubgen_datetime_datetime
        self.nicknames: DataFrame
        self.topping: __stubgen_tests_features_test__df_Topping
        self.foo: __stubgen_tests_features_test__df_Foo

class TacoProtocol(Protocol):
    id: str
    user_id: str
    price: int
    maybe_price: Union[int, None]
    hat: str
    topping_id: str
    ts: __stubgen_datetime_datetime
    nicknames: DataFrame
    topping: __stubgen_tests_features_test__df_Topping
    foo: __stubgen_tests_features_test__df_Foo

class HelloMetaclass(FeaturesMeta):
    @property
    def a(self) -> Type[str]: ...

    @property
    def b(self) -> Type[int]: ...

    @property
    def c(self) -> Type[__stubgen_datetime_datetime]: ...

    @property
    def d(self) -> Type[__stubgen_tests_serialization_test__codec_Color]: ...

    @property
    def e(self) -> Type[__stubgen_datetime_date]: ...

    @property
    def f(self) -> Type[__stubgen_tests_serialization_test__codec_MyDataclass]: ...

    @property
    def y(self) -> Type[set[int]]: ...

    @property
    def z(self) -> Type[list[str]]: ...

    @property
    def fancy(self) -> Type[__stubgen_tests_serialization_test__codec_CustomClass]: ...

class Hello(Features, metaclass=HelloMetaclass):
    def __init__(
        self,
        a: str = ...,
        b: int = ...,
        c: __stubgen_datetime_datetime = ...,
        d: __stubgen_tests_serialization_test__codec_Color = ...,
        e: __stubgen_datetime_date = ...,
        f: __stubgen_tests_serialization_test__codec_MyDataclass = ...,
        y: set[int] = ...,
        z: list[str] = ...,
        fancy: __stubgen_tests_serialization_test__codec_CustomClass = ...,
    ):
        self.a: str
        self.b: int
        self.c: __stubgen_datetime_datetime
        self.d: __stubgen_tests_serialization_test__codec_Color
        self.e: __stubgen_datetime_date
        self.f: __stubgen_tests_serialization_test__codec_MyDataclass
        self.y: set[int]
        self.z: list[str]
        self.fancy: __stubgen_tests_serialization_test__codec_CustomClass

class HelloProtocol(Protocol):
    a: str
    b: int
    c: __stubgen_datetime_datetime
    d: __stubgen_tests_serialization_test__codec_Color
    e: __stubgen_datetime_date
    f: __stubgen_tests_serialization_test__codec_MyDataclass
    y: set[int]
    z: list[str]
    fancy: __stubgen_tests_serialization_test__codec_CustomClass

class CommentBaseOwnerMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def empty(self) -> Type[str]: ...

    @property
    def email(self) -> Type[str]: ...

    @property
    def email_commas(self) -> Type[str]: ...

    @property
    def email_single(self) -> Type[str]: ...

    @property
    def email_all_kinds(self) -> Type[str]: ...

class CommentBaseOwner(Features, metaclass=CommentBaseOwnerMetaclass):
    def __init__(
        self,
        id: str = ...,
        empty: str = ...,
        email: str = ...,
        email_commas: str = ...,
        email_single: str = ...,
        email_all_kinds: str = ...,
    ):
        self.id: str
        self.empty: str
        self.email: str
        self.email_commas: str
        self.email_single: str
        self.email_all_kinds: str

class CommentBaseOwnerProtocol(Protocol):
    id: str
    empty: str
    email: str
    email_commas: str
    email_single: str
    email_all_kinds: str

class FootMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def is_flat(self) -> Type[bool]: ...

    @property
    def person_id(self) -> Type[int]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

    @property
    def ankle(self) -> Type[__stubgen_tests_client_test__expand__features_Ankle]: ...

class Foot(Features, metaclass=FootMetaclass):
    def __init__(
        self,
        id: int = ...,
        is_flat: bool = ...,
        person_id: int = ...,
        ts: __stubgen_datetime_datetime = ...,
        ankle: __stubgen_tests_client_test__expand__features_Ankle = ...,
    ):
        self.id: int
        self.is_flat: bool
        self.person_id: int
        self.ts: __stubgen_datetime_datetime
        self.ankle: __stubgen_tests_client_test__expand__features_Ankle

class FootProtocol(Protocol):
    id: int
    is_flat: bool
    person_id: int
    ts: __stubgen_datetime_datetime
    ankle: __stubgen_tests_client_test__expand__features_Ankle

class PersonMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def name(self) -> Type[str]: ...

    @property
    def best_foot_id(self) -> Type[int]: ...

    @property
    def best_foot(self) -> Type[__stubgen_tests_client_test__expand__features_Foot]: ...

    @property
    def feet(self) -> Type[DataFrame]: ...

class Person(Features, metaclass=PersonMetaclass):
    def __init__(
        self,
        id: int = ...,
        name: str = ...,
        best_foot_id: int = ...,
        best_foot: __stubgen_tests_client_test__expand__features_Foot = ...,
        feet: DataFrame = ...,
    ):
        self.id: int
        self.name: str
        self.best_foot_id: int
        self.best_foot: __stubgen_tests_client_test__expand__features_Foot
        self.feet: DataFrame

class PersonProtocol(Protocol):
    id: int
    name: str
    best_foot_id: int
    best_foot: __stubgen_tests_client_test__expand__features_Foot
    feet: DataFrame

class HomeFeaturesChainedFeatureTimeMetaclass(FeaturesMeta):
    @property
    def home_id(self) -> Type[str]: ...

    @property
    def address(self) -> Type[str]: ...

    @property
    def price(self) -> Type[int]: ...

    @property
    def sq_ft(self) -> Type[int]: ...

    @property
    def homeowner(self) -> Type[__stubgen_tests_features_test__chained__feature__time_Homeowner]: ...

class HomeFeaturesChainedFeatureTime(Features, metaclass=HomeFeaturesChainedFeatureTimeMetaclass):
    def __init__(
        self,
        home_id: str = ...,
        address: str = ...,
        price: int = ...,
        sq_ft: int = ...,
        homeowner: __stubgen_tests_features_test__chained__feature__time_Homeowner = ...,
    ):
        self.home_id: str
        self.address: str
        self.price: int
        self.sq_ft: int
        self.homeowner: __stubgen_tests_features_test__chained__feature__time_Homeowner

class HomeFeaturesChainedFeatureTimeProtocol(Protocol):
    home_id: str
    address: str
    price: int
    sq_ft: int
    homeowner: __stubgen_tests_features_test__chained__feature__time_Homeowner

class ToppingMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def value(self) -> Type[int]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

    @property
    def nicknames(self) -> Type[DataFrame]: ...

    @property
    def price(self) -> Type[__stubgen_tests_features_test__df_ToppingPrice]: ...

class Topping(Features, metaclass=ToppingMetaclass):
    def __init__(
        self,
        id: str = ...,
        value: int = ...,
        ts: __stubgen_datetime_datetime = ...,
        nicknames: DataFrame = ...,
        price: __stubgen_tests_features_test__df_ToppingPrice = ...,
    ):
        self.id: str
        self.value: int
        self.ts: __stubgen_datetime_datetime
        self.nicknames: DataFrame
        self.price: __stubgen_tests_features_test__df_ToppingPrice

class ToppingProtocol(Protocol):
    id: str
    value: int
    ts: __stubgen_datetime_datetime
    nicknames: DataFrame
    price: __stubgen_tests_features_test__df_ToppingPrice

class ToppingPriceMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def topping_id(self) -> Type[str]: ...

    @property
    def wow(self) -> Type[str]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

class ToppingPrice(Features, metaclass=ToppingPriceMetaclass):
    def __init__(
        self,
        id: str = ...,
        topping_id: str = ...,
        wow: str = ...,
        ts: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.topping_id: str
        self.wow: str
        self.ts: __stubgen_datetime_datetime

class ToppingPriceProtocol(Protocol):
    id: str
    topping_id: str
    wow: str
    ts: __stubgen_datetime_datetime

class OwnerFeaturesMetaclass(FeaturesMeta):
    @property
    def plain(self) -> Type[str]: ...

    @property
    def cached(self) -> Type[str]: ...

    @property
    def andy(self) -> Type[str]: ...

    @property
    def ft(self) -> Type[__stubgen_datetime_datetime]: ...

class OwnerFeatures(Features, metaclass=OwnerFeaturesMetaclass):
    def __init__(
        self,
        plain: str = ...,
        cached: str = ...,
        andy: str = ...,
        ft: __stubgen_datetime_datetime = ...,
    ):
        self.plain: str
        self.cached: str
        self.andy: str
        self.ft: __stubgen_datetime_datetime

class OwnerFeaturesProtocol(Protocol):
    plain: str
    cached: str
    andy: str
    ft: __stubgen_datetime_datetime

class TagFeaturesMetaclass(FeaturesMeta):
    @property
    def empty(self) -> Type[str]: ...

    @property
    def one(self) -> Type[str]: ...

    @property
    def many(self) -> Type[str]: ...

    @property
    def ft(self) -> Type[__stubgen_datetime_datetime]: ...

class TagFeatures(Features, metaclass=TagFeaturesMetaclass):
    def __init__(
        self,
        empty: str = ...,
        one: str = ...,
        many: str = ...,
        ft: __stubgen_datetime_datetime = ...,
    ):
        self.empty: str
        self.one: str
        self.many: str
        self.ft: __stubgen_datetime_datetime

class TagFeaturesProtocol(Protocol):
    empty: str
    one: str
    many: str
    ft: __stubgen_datetime_datetime

class ParentFSMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def children(self) -> Type[DataFrame]: ...

    @property
    def single_child(self) -> Type[__stubgen_tests_features_test__features_SingleChildFS]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

class ParentFS(Features, metaclass=ParentFSMetaclass):
    def __init__(
        self,
        id: str = ...,
        children: DataFrame = ...,
        single_child: __stubgen_tests_features_test__features_SingleChildFS = ...,
        ts: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.children: DataFrame
        self.single_child: __stubgen_tests_features_test__features_SingleChildFS
        self.ts: __stubgen_datetime_datetime

class ParentFSProtocol(Protocol):
    id: str
    children: DataFrame
    single_child: __stubgen_tests_features_test__features_SingleChildFS
    ts: __stubgen_datetime_datetime

class FunFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def nope(self) -> Type[str]: ...

    @property
    def single_parent(self) -> Type[__stubgen_tests_features_test__iter_NoFunFeatures]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

class FunFeatures(Features, metaclass=FunFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        nope: str = ...,
        single_parent: __stubgen_tests_features_test__iter_NoFunFeatures = ...,
        ts: __stubgen_datetime_datetime = ...,
    ):
        self.id: str
        self.nope: str
        self.single_parent: __stubgen_tests_features_test__iter_NoFunFeatures
        self.ts: __stubgen_datetime_datetime

class FunFeaturesProtocol(Protocol):
    id: str
    nope: str
    single_parent: __stubgen_tests_features_test__iter_NoFunFeatures
    ts: __stubgen_datetime_datetime

class HomeFeaturesMetaclass(FeaturesMeta):
    @property
    def home_id(self) -> Type[str]: ...

    @property
    def address(self) -> Type[str]: ...

    @property
    def price(self) -> Type[int]: ...

    @property
    def sq_ft(self) -> Type[int]: ...

class HomeFeatures(Features, metaclass=HomeFeaturesMetaclass):
    def __init__(
        self,
        home_id: str = ...,
        address: str = ...,
        price: int = ...,
        sq_ft: int = ...,
    ):
        self.home_id: str
        self.address: str
        self.price: int
        self.sq_ft: int

class HomeFeaturesProtocol(Protocol):
    home_id: str
    address: str
    price: int
    sq_ft: int

class TransactionMetaclass(FeaturesMeta):
    @property
    def user_id(self) -> Type[str]: ...

    @property
    def amount(self) -> Type[int]: ...

    @property
    def source_account_id(self) -> Type[int]: ...

    @property
    def dest_account_id(self) -> Type[int]: ...

class Transaction(Features, metaclass=TransactionMetaclass):
    def __init__(
        self,
        user_id: str = ...,
        amount: int = ...,
        source_account_id: int = ...,
        dest_account_id: int = ...,
    ):
        self.user_id: str
        self.amount: int
        self.source_account_id: int
        self.dest_account_id: int

class TransactionProtocol(Protocol):
    user_id: str
    amount: int
    source_account_id: int
    dest_account_id: int

class MypyUserFeaturesMetaclass(FeaturesMeta):
    @property
    def uid(self) -> Type[str]: ...

    @property
    def name(self) -> Type[str]: ...

    @property
    def bday(self) -> Type[str]: ...

    @property
    def age(self) -> Type[int]: ...

class MypyUserFeatures(Features, metaclass=MypyUserFeaturesMetaclass):
    def __init__(
        self,
        uid: str = ...,
        name: str = ...,
        bday: str = ...,
        age: int = ...,
    ):
        self.uid: str
        self.name: str
        self.bday: str
        self.age: int

class MypyUserFeaturesProtocol(Protocol):
    uid: str
    name: str
    bday: str
    age: int

class UserMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def fav_color(self) -> Type[__stubgen_tests_client_test__client__serialization_Color]: ...

    @property
    def birthday(self) -> Type[__stubgen_datetime_date]: ...

class User(Features, metaclass=UserMetaclass):
    def __init__(
        self,
        id: int = ...,
        fav_color: __stubgen_tests_client_test__client__serialization_Color = ...,
        birthday: __stubgen_datetime_date = ...,
    ):
        self.id: int
        self.fav_color: __stubgen_tests_client_test__client__serialization_Color
        self.birthday: __stubgen_datetime_date

class UserProtocol(Protocol):
    id: int
    fav_color: __stubgen_tests_client_test__client__serialization_Color
    birthday: __stubgen_datetime_date

class HomeownerMetaclass(FeaturesMeta):
    @property
    def fullname(self) -> Type[str]: ...

    @property
    def home_id(self) -> Type[str]: ...

    @property
    def ts(self) -> Type[__stubgen_datetime_datetime]: ...

class Homeowner(Features, metaclass=HomeownerMetaclass):
    def __init__(
        self,
        fullname: str = ...,
        home_id: str = ...,
        ts: __stubgen_datetime_datetime = ...,
    ):
        self.fullname: str
        self.home_id: str
        self.ts: __stubgen_datetime_datetime

class HomeownerProtocol(Protocol):
    fullname: str
    home_id: str
    ts: __stubgen_datetime_datetime

class MaxStalenessFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def name(self) -> Type[str]: ...

    @property
    def woohoo(self) -> Type[str]: ...

class MaxStalenessFeatures(Features, metaclass=MaxStalenessFeaturesMetaclass):
    def __init__(
        self,
        id: int = ...,
        name: str = ...,
        woohoo: str = ...,
    ):
        self.id: int
        self.name: str
        self.woohoo: str

class MaxStalenessFeaturesProtocol(Protocol):
    id: int
    name: str
    woohoo: str

class ETLFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def name(self) -> Type[str]: ...

    @property
    def woohoo(self) -> Type[str]: ...

class ETLFeatures(Features, metaclass=ETLFeaturesMetaclass):
    def __init__(
        self,
        id: int = ...,
        name: str = ...,
        woohoo: str = ...,
    ):
        self.id: int
        self.name: str
        self.woohoo: str

class ETLFeaturesProtocol(Protocol):
    id: int
    name: str
    woohoo: str

class ChildFSMetaclass(FeaturesMeta):
    @property
    def parent_id(self) -> Type[str]: ...

    @property
    def parents(self) -> Type[DataFrame]: ...

    @property
    def single_parent(self) -> Type[__stubgen_tests_features_test__features_SingleParentFS]: ...

class ChildFS(Features, metaclass=ChildFSMetaclass):
    def __init__(
        self,
        parent_id: str = ...,
        parents: DataFrame = ...,
        single_parent: __stubgen_tests_features_test__features_SingleParentFS = ...,
    ):
        self.parent_id: str
        self.parents: DataFrame
        self.single_parent: __stubgen_tests_features_test__features_SingleParentFS

class ChildFSProtocol(Protocol):
    parent_id: str
    parents: DataFrame
    single_parent: __stubgen_tests_features_test__features_SingleParentFS

class OtherPrimaryKeyFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def nah_really_this_is_id(self) -> Type[str]: ...

    @property
    def other(self) -> Type[str]: ...

class OtherPrimaryKeyFeatures(Features, metaclass=OtherPrimaryKeyFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        nah_really_this_is_id: str = ...,
        other: str = ...,
    ):
        self.id: str
        self.nah_really_this_is_id: str
        self.other: str

class OtherPrimaryKeyFeaturesProtocol(Protocol):
    id: str
    nah_really_this_is_id: str
    other: str

class UserProfileMetaclass(FeaturesMeta):
    @property
    def user_id(self) -> Type[str]: ...

    @property
    def profile_id(self) -> Type[str]: ...

    @property
    def address(self) -> Type[int]: ...

class UserProfile(Features, metaclass=UserProfileMetaclass):
    def __init__(
        self,
        user_id: str = ...,
        profile_id: str = ...,
        address: int = ...,
    ):
        self.user_id: str
        self.profile_id: str
        self.address: int

class UserProfileProtocol(Protocol):
    user_id: str
    profile_id: str
    address: int

class AnkleMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[int]: ...

    @property
    def foot_id(self) -> Type[int]: ...

class Ankle(Features, metaclass=AnkleMetaclass):
    def __init__(
        self,
        id: int = ...,
        foot_id: int = ...,
    ):
        self.id: int
        self.foot_id: int

class AnkleProtocol(Protocol):
    id: int
    foot_id: int

class ExampleFraudOrgMetaclass(FeaturesMeta):
    @property
    def uid(self) -> Type[str]: ...

    @property
    def org_name(self) -> Type[str]: ...

class ExampleFraudOrg(Features, metaclass=ExampleFraudOrgMetaclass):
    def __init__(
        self,
        uid: str = ...,
        org_name: str = ...,
    ):
        self.uid: str
        self.org_name: str

class ExampleFraudOrgProtocol(Protocol):
    uid: str
    org_name: str

class ExampleFraudUserMetaclass(FeaturesMeta):
    @property
    def uid(self) -> Type[str]: ...

    @property
    def org(self) -> Type[__stubgen_tests_features_test__chained__has__one_ExampleFraudOrg]: ...

class ExampleFraudUser(Features, metaclass=ExampleFraudUserMetaclass):
    def __init__(
        self,
        uid: str = ...,
        org: __stubgen_tests_features_test__chained__has__one_ExampleFraudOrg = ...,
    ):
        self.uid: str
        self.org: __stubgen_tests_features_test__chained__has__one_ExampleFraudOrg

class ExampleFraudUserProtocol(Protocol):
    uid: str
    org: __stubgen_tests_features_test__chained__has__one_ExampleFraudOrg

class ExampleFraudProfileMetaclass(FeaturesMeta):
    @property
    def uid(self) -> Type[str]: ...

    @property
    def user(self) -> Type[__stubgen_tests_features_test__chained__has__one_ExampleFraudUser]: ...

class ExampleFraudProfile(Features, metaclass=ExampleFraudProfileMetaclass):
    def __init__(
        self,
        uid: str = ...,
        user: __stubgen_tests_features_test__chained__has__one_ExampleFraudUser = ...,
    ):
        self.uid: str
        self.user: __stubgen_tests_features_test__chained__has__one_ExampleFraudUser

class ExampleFraudProfileProtocol(Protocol):
    uid: str
    user: __stubgen_tests_features_test__chained__has__one_ExampleFraudUser

class CustomNameClassMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def other(self) -> Type[str]: ...

class CustomNameClass(Features, metaclass=CustomNameClassMetaclass):
    def __init__(
        self,
        id: str = ...,
        other: str = ...,
    ):
        self.id: str
        self.other: str

class CustomNameClassProtocol(Protocol):
    id: str
    other: str

class SingleChildFSMetaclass(FeaturesMeta):
    @property
    def parent_id(self) -> Type[str]: ...

    @property
    def parent(self) -> Type[DataFrame]: ...

class SingleChildFS(Features, metaclass=SingleChildFSMetaclass):
    def __init__(
        self,
        parent_id: str = ...,
        parent: DataFrame = ...,
    ):
        self.parent_id: str
        self.parent: DataFrame

class SingleChildFSProtocol(Protocol):
    parent_id: str
    parent: DataFrame

class SingleParentFSMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def children(self) -> Type[DataFrame]: ...

class SingleParentFS(Features, metaclass=SingleParentFSMetaclass):
    def __init__(
        self,
        id: str = ...,
        children: DataFrame = ...,
    ):
        self.id: str
        self.children: DataFrame

class SingleParentFSProtocol(Protocol):
    id: str
    children: DataFrame

class UnassignedIdFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def other(self) -> Type[str]: ...

class UnassignedIdFeatures(Features, metaclass=UnassignedIdFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        other: str = ...,
    ):
        self.id: str
        self.other: str

class UnassignedIdFeaturesProtocol(Protocol):
    id: str
    other: str

class UnassignedDecoratedIdFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def other(self) -> Type[str]: ...

class UnassignedDecoratedIdFeatures(Features, metaclass=UnassignedDecoratedIdFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        other: str = ...,
    ):
        self.id: str
        self.other: str

class UnassignedDecoratedIdFeaturesProtocol(Protocol):
    id: str
    other: str

class ExplicitIdFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def other(self) -> Type[str]: ...

class ExplicitIdFeatures(Features, metaclass=ExplicitIdFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        other: str = ...,
    ):
        self.id: str
        self.other: str

class ExplicitIdFeaturesProtocol(Protocol):
    id: str
    other: str

class BogusIdFeature1Metaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def other(self) -> Type[str]: ...

class BogusIdFeature1(Features, metaclass=BogusIdFeature1Metaclass):
    def __init__(
        self,
        id: str = ...,
        other: str = ...,
    ):
        self.id: str
        self.other: str

class BogusIdFeature1Protocol(Protocol):
    id: str
    other: str

class BogusIdFeature2Metaclass(FeaturesMeta):
    @property
    def id(self) -> Type[DataFrame]: ...

    @property
    def other(self) -> Type[str]: ...

class BogusIdFeature2(Features, metaclass=BogusIdFeature2Metaclass):
    def __init__(
        self,
        id: DataFrame = ...,
        other: str = ...,
    ):
        self.id: DataFrame
        self.other: str

class BogusIdFeature2Protocol(Protocol):
    id: DataFrame
    other: str

class BogusIdFeature3Metaclass(FeaturesMeta):
    @property
    def id(self) -> Type[__stubgen_datetime_datetime]: ...

    @property
    def other(self) -> Type[str]: ...

class BogusIdFeature3(Features, metaclass=BogusIdFeature3Metaclass):
    def __init__(
        self,
        id: __stubgen_datetime_datetime = ...,
        other: str = ...,
    ):
        self.id: __stubgen_datetime_datetime
        self.other: str

class BogusIdFeature3Protocol(Protocol):
    id: __stubgen_datetime_datetime
    other: str

class IdIsNotIdFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def other(self) -> Type[str]: ...

class IdIsNotIdFeatures(Features, metaclass=IdIsNotIdFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
        other: str = ...,
    ):
        self.id: str
        self.other: str

class IdIsNotIdFeaturesProtocol(Protocol):
    id: str
    other: str

class NotIdIsIdFeaturesMetaclass(FeaturesMeta):
    @property
    def not_id(self) -> Type[str]: ...

    @property
    def other(self) -> Type[str]: ...

class NotIdIsIdFeatures(Features, metaclass=NotIdIsIdFeaturesMetaclass):
    def __init__(
        self,
        not_id: str = ...,
        other: str = ...,
    ):
        self.not_id: str
        self.other: str

class NotIdIsIdFeaturesProtocol(Protocol):
    not_id: str
    other: str

class AcctMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

    @property
    def balance(self) -> Type[float]: ...

class Acct(Features, metaclass=AcctMetaclass):
    def __init__(
        self,
        id: str = ...,
        balance: float = ...,
    ):
        self.id: str
        self.balance: float

class AcctProtocol(Protocol):
    id: str
    balance: float

class AccountMetaclass(FeaturesMeta):
    @property
    def account_id(self) -> Type[str]: ...

    @property
    def balance(self) -> Type[int]: ...

class Account(Features, metaclass=AccountMetaclass):
    def __init__(
        self,
        account_id: str = ...,
        balance: int = ...,
    ):
        self.account_id: str
        self.balance: int

class AccountProtocol(Protocol):
    account_id: str
    balance: int

class StreamFeaturesWindowMetaclass(FeaturesMeta):
    @property
    def scalar_feature(self) -> Type[__stubgen_chalk_streams___windows_Windowed]: ...

    @property
    def uid(self) -> Type[str]: ...

class StreamFeaturesWindow(Features, metaclass=StreamFeaturesWindowMetaclass):
    def __init__(
        self,
        scalar_feature: __stubgen_chalk_streams___windows_Windowed = ...,
        uid: str = ...,
    ):
        self.scalar_feature: __stubgen_chalk_streams___windows_Windowed
        self.uid: str

class StreamFeaturesWindowProtocol(Protocol):
    scalar_feature: __stubgen_chalk_streams___windows_Windowed
    uid: str

class NicknameMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

class Nickname(Features, metaclass=NicknameMetaclass):
    def __init__(
        self,
        id: str = ...,
    ):
        self.id: str

class NicknameProtocol(Protocol):
    id: str

class NoFunFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

class NoFunFeatures(Features, metaclass=NoFunFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
    ):
        self.id: str

class NoFunFeaturesProtocol(Protocol):
    id: str

class UserFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

class UserFeatures(Features, metaclass=UserFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
    ):
        self.id: str

class UserFeaturesProtocol(Protocol):
    id: str

class LibraryFeaturesMetaclass(FeaturesMeta):
    @property
    def id(self) -> Type[str]: ...

class LibraryFeatures(Features, metaclass=LibraryFeaturesMetaclass):
    def __init__(
        self,
        id: str = ...,
    ):
        self.id: str

class LibraryFeaturesProtocol(Protocol):
    id: str

class StreamFeaturesMetaclass(FeaturesMeta):
    @property
    def scalar_feature(self) -> Type[str]: ...

class StreamFeatures(Features, metaclass=StreamFeaturesMetaclass):
    def __init__(
        self,
        scalar_feature: str = ...,
    ):
        self.scalar_feature: str

class StreamFeaturesProtocol(Protocol):
    scalar_feature: str

@overload
def features(item: Type[WowFSProtocol]) -> Type[WowFS]: ...

@overload
def features(item: Type[TacoProtocol]) -> Type[Taco]: ...

@overload
def features(item: Type[HelloProtocol]) -> Type[Hello]: ...

@overload
def features(item: Type[CommentBaseOwnerProtocol]) -> Type[CommentBaseOwner]: ...

@overload
def features(item: Type[FootProtocol]) -> Type[Foot]: ...

@overload
def features(item: Type[PersonProtocol]) -> Type[Person]: ...

@overload
def features(item: Type[HomeFeaturesChainedFeatureTimeProtocol]) -> Type[HomeFeaturesChainedFeatureTime]: ...

@overload
def features(item: Type[ToppingProtocol]) -> Type[Topping]: ...

@overload
def features(item: Type[ToppingPriceProtocol]) -> Type[ToppingPrice]: ...

@overload
def features(item: Type[OwnerFeaturesProtocol]) -> Type[OwnerFeatures]: ...

@overload
def features(item: Type[TagFeaturesProtocol]) -> Type[TagFeatures]: ...

@overload
def features(item: Type[ParentFSProtocol]) -> Type[ParentFS]: ...

@overload
def features(item: Type[FunFeaturesProtocol]) -> Type[FunFeatures]: ...

@overload
def features(item: Type[HomeFeaturesProtocol]) -> Type[HomeFeatures]: ...

@overload
def features(item: Type[TransactionProtocol]) -> Type[Transaction]: ...

@overload
def features(item: Type[MypyUserFeaturesProtocol]) -> Type[MypyUserFeatures]: ...

@overload
def features(item: Type[UserProtocol]) -> Type[User]: ...

@overload
def features(item: Type[HomeownerProtocol]) -> Type[Homeowner]: ...

@overload
def features(item: Type[MaxStalenessFeaturesProtocol]) -> Type[MaxStalenessFeatures]: ...

@overload
def features(item: Type[ETLFeaturesProtocol]) -> Type[ETLFeatures]: ...

@overload
def features(item: Type[ChildFSProtocol]) -> Type[ChildFS]: ...

@overload
def features(item: Type[OtherPrimaryKeyFeaturesProtocol]) -> Type[OtherPrimaryKeyFeatures]: ...

@overload
def features(item: Type[UserProfileProtocol]) -> Type[UserProfile]: ...

@overload
def features(item: Type[AnkleProtocol]) -> Type[Ankle]: ...

@overload
def features(item: Type[ExampleFraudOrgProtocol]) -> Type[ExampleFraudOrg]: ...

@overload
def features(item: Type[ExampleFraudUserProtocol]) -> Type[ExampleFraudUser]: ...

@overload
def features(item: Type[ExampleFraudProfileProtocol]) -> Type[ExampleFraudProfile]: ...

@overload
def features(item: Type[CustomNameClassProtocol]) -> Type[CustomNameClass]: ...

@overload
def features(item: Type[SingleChildFSProtocol]) -> Type[SingleChildFS]: ...

@overload
def features(item: Type[SingleParentFSProtocol]) -> Type[SingleParentFS]: ...

@overload
def features(item: Type[UnassignedIdFeaturesProtocol]) -> Type[UnassignedIdFeatures]: ...

@overload
def features(item: Type[UnassignedDecoratedIdFeaturesProtocol]) -> Type[UnassignedDecoratedIdFeatures]: ...

@overload
def features(item: Type[ExplicitIdFeaturesProtocol]) -> Type[ExplicitIdFeatures]: ...

@overload
def features(item: Type[BogusIdFeature1Protocol]) -> Type[BogusIdFeature1]: ...

@overload
def features(item: Type[BogusIdFeature2Protocol]) -> Type[BogusIdFeature2]: ...

@overload
def features(item: Type[BogusIdFeature3Protocol]) -> Type[BogusIdFeature3]: ...

@overload
def features(item: Type[IdIsNotIdFeaturesProtocol]) -> Type[IdIsNotIdFeatures]: ...

@overload
def features(item: Type[NotIdIsIdFeaturesProtocol]) -> Type[NotIdIsIdFeatures]: ...

@overload
def features(item: Type[AcctProtocol]) -> Type[Acct]: ...

@overload
def features(item: Type[AccountProtocol]) -> Type[Account]: ...

@overload
def features(item: Type[StreamFeaturesWindowProtocol]) -> Type[StreamFeaturesWindow]: ...

@overload
def features(item: Type[NicknameProtocol]) -> Type[Nickname]: ...

@overload
def features(item: Type[NoFunFeaturesProtocol]) -> Type[NoFunFeatures]: ...

@overload
def features(item: Type[UserFeaturesProtocol]) -> Type[UserFeatures]: ...

@overload
def features(item: Type[LibraryFeaturesProtocol]) -> Type[LibraryFeatures]: ...

@overload
def features(item: Type[StreamFeaturesProtocol]) -> Type[StreamFeatures]: ...

@overload
def features(
    *,
    owner: Optional[str] = ...,
    tags: Optional[Tags] = ...,
    max_staleness: Optional[Duration] = ...,
    etl_offline_to_online: Optional[bool] = ...,
) -> __stubgen__features_proto: ...

class __stubgen__features_proto(Protocol):
    @overload
    def __call__(self, item: Type[WowFSProtocol]) -> Type[WowFS]: ...

    @overload
    def __call__(self, item: Type[TacoProtocol]) -> Type[Taco]: ...

    @overload
    def __call__(self, item: Type[HelloProtocol]) -> Type[Hello]: ...

    @overload
    def __call__(self, item: Type[CommentBaseOwnerProtocol]) -> Type[CommentBaseOwner]: ...

    @overload
    def __call__(self, item: Type[FootProtocol]) -> Type[Foot]: ...

    @overload
    def __call__(self, item: Type[PersonProtocol]) -> Type[Person]: ...

    @overload
    def __call__(self, item: Type[HomeFeaturesChainedFeatureTimeProtocol]) -> Type[HomeFeaturesChainedFeatureTime]: ...

    @overload
    def __call__(self, item: Type[ToppingProtocol]) -> Type[Topping]: ...

    @overload
    def __call__(self, item: Type[ToppingPriceProtocol]) -> Type[ToppingPrice]: ...

    @overload
    def __call__(self, item: Type[OwnerFeaturesProtocol]) -> Type[OwnerFeatures]: ...

    @overload
    def __call__(self, item: Type[TagFeaturesProtocol]) -> Type[TagFeatures]: ...

    @overload
    def __call__(self, item: Type[ParentFSProtocol]) -> Type[ParentFS]: ...

    @overload
    def __call__(self, item: Type[FunFeaturesProtocol]) -> Type[FunFeatures]: ...

    @overload
    def __call__(self, item: Type[HomeFeaturesProtocol]) -> Type[HomeFeatures]: ...

    @overload
    def __call__(self, item: Type[TransactionProtocol]) -> Type[Transaction]: ...

    @overload
    def __call__(self, item: Type[MypyUserFeaturesProtocol]) -> Type[MypyUserFeatures]: ...

    @overload
    def __call__(self, item: Type[UserProtocol]) -> Type[User]: ...

    @overload
    def __call__(self, item: Type[HomeownerProtocol]) -> Type[Homeowner]: ...

    @overload
    def __call__(self, item: Type[MaxStalenessFeaturesProtocol]) -> Type[MaxStalenessFeatures]: ...

    @overload
    def __call__(self, item: Type[ETLFeaturesProtocol]) -> Type[ETLFeatures]: ...

    @overload
    def __call__(self, item: Type[ChildFSProtocol]) -> Type[ChildFS]: ...

    @overload
    def __call__(self, item: Type[OtherPrimaryKeyFeaturesProtocol]) -> Type[OtherPrimaryKeyFeatures]: ...

    @overload
    def __call__(self, item: Type[UserProfileProtocol]) -> Type[UserProfile]: ...

    @overload
    def __call__(self, item: Type[AnkleProtocol]) -> Type[Ankle]: ...

    @overload
    def __call__(self, item: Type[ExampleFraudOrgProtocol]) -> Type[ExampleFraudOrg]: ...

    @overload
    def __call__(self, item: Type[ExampleFraudUserProtocol]) -> Type[ExampleFraudUser]: ...

    @overload
    def __call__(self, item: Type[ExampleFraudProfileProtocol]) -> Type[ExampleFraudProfile]: ...

    @overload
    def __call__(self, item: Type[CustomNameClassProtocol]) -> Type[CustomNameClass]: ...

    @overload
    def __call__(self, item: Type[SingleChildFSProtocol]) -> Type[SingleChildFS]: ...

    @overload
    def __call__(self, item: Type[SingleParentFSProtocol]) -> Type[SingleParentFS]: ...

    @overload
    def __call__(self, item: Type[UnassignedIdFeaturesProtocol]) -> Type[UnassignedIdFeatures]: ...

    @overload
    def __call__(self, item: Type[UnassignedDecoratedIdFeaturesProtocol]) -> Type[UnassignedDecoratedIdFeatures]: ...

    @overload
    def __call__(self, item: Type[ExplicitIdFeaturesProtocol]) -> Type[ExplicitIdFeatures]: ...

    @overload
    def __call__(self, item: Type[BogusIdFeature1Protocol]) -> Type[BogusIdFeature1]: ...

    @overload
    def __call__(self, item: Type[BogusIdFeature2Protocol]) -> Type[BogusIdFeature2]: ...

    @overload
    def __call__(self, item: Type[BogusIdFeature3Protocol]) -> Type[BogusIdFeature3]: ...

    @overload
    def __call__(self, item: Type[IdIsNotIdFeaturesProtocol]) -> Type[IdIsNotIdFeatures]: ...

    @overload
    def __call__(self, item: Type[NotIdIsIdFeaturesProtocol]) -> Type[NotIdIsIdFeatures]: ...

    @overload
    def __call__(self, item: Type[AcctProtocol]) -> Type[Acct]: ...

    @overload
    def __call__(self, item: Type[AccountProtocol]) -> Type[Account]: ...

    @overload
    def __call__(self, item: Type[StreamFeaturesWindowProtocol]) -> Type[StreamFeaturesWindow]: ...

    @overload
    def __call__(self, item: Type[NicknameProtocol]) -> Type[Nickname]: ...

    @overload
    def __call__(self, item: Type[NoFunFeaturesProtocol]) -> Type[NoFunFeatures]: ...

    @overload
    def __call__(self, item: Type[UserFeaturesProtocol]) -> Type[UserFeatures]: ...

    @overload
    def __call__(self, item: Type[LibraryFeaturesProtocol]) -> Type[LibraryFeatures]: ...

    @overload
    def __call__(self, item: Type[StreamFeaturesProtocol]) -> Type[StreamFeatures]: ...



import logging
import time
import typing
from collections import defaultdict
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Mapping, Optional, Union

import sqlalchemy
from dateutil import parser
from polars import from_arrow
from sqlalchemy import text
from sqlalchemy.engine.row import Row

from chalk.features import DataFrame, Feature, FeatureSetBase
from chalk.sql.base.protocols import (
    BaseSQLSourceProtocol,
    ChalkQueryProtocol,
    DBSessionProtocol,
    IncrementalSettings,
    StringChalkQueryProtocol,
)
from chalk.utils.collections import get_unique_item


class Finalizer(str, Enum):
    OneOrNone = "OneOrNone"
    One = "One"
    First = "First"
    All = "All"


_logger = logging.getLogger(__name__)


def _maybe_coerce_feature_value(value: Any, feature: Feature, coerce: bool):
    if not coerce:
        return value

    if value is None:
        return value

    assert feature.typ is not None

    if issubclass(feature.typ.underlying, datetime) and isinstance(value, str):
        # Some SQL dialects/drivers, such as SQLite, will return datetimes as raw strings
        # TODO: there's no guarantee that this `value` is actually an ISO string. SQLite stores and returns them
        # in the format "%Y-%m-%d %H:%M:%S", which seems good enough for parser.isoparse()
        return parser.isoparse(value)

    return value


def _construct_features(
    cols: Union[List[Feature], Dict[str, Feature]], tuples: Optional[Row], coerce_raw_values: bool = False
):
    root_ns = get_unique_item(
        (f.root_namespace for f in (cols.values() if isinstance(cols, dict) else cols)), "root ns"
    )
    features_cls = FeatureSetBase.registry[root_ns]
    kwargs: Dict[str, Any] = {}

    if isinstance(cols, dict):
        if tuples is None:
            for k, feature in cols.items():
                kwargs[feature.attribute_name] = None
        else:
            row = dict(tuples)
            for k, feature in cols.items():
                assert feature.attribute_name is not None
                kwargs[feature.attribute_name] = _maybe_coerce_feature_value(
                    row[k] if row else None, feature, coerce_raw_values
                )
    else:
        assert tuples is not None
        assert len(tuples) == len(cols)
        for feature, val in zip(cols, tuples):
            assert feature.attribute_name is not None
            kwargs[feature.attribute_name] = _maybe_coerce_feature_value(val, feature, coerce_raw_values)
    return features_cls(**kwargs)


class StringChalkQuery(StringChalkQueryProtocol):
    def __init__(
        self,
        session: DBSessionProtocol,
        source: BaseSQLSourceProtocol,
        query: str,
        fields: Mapping[str, Union[Feature, str]],
        args: Optional[Mapping[str, str]],
    ):
        if len(fields) == 0:
            raise ValueError("Expected argument 'fields' to be a non-empty dict, but got {}")

        self._finalizer: Optional[Finalizer] = None
        self._session = session
        self._source = source
        self._original_query = query
        self._query = text(query)
        self._fields = fields
        self._incremental_settings: Optional[Union[IncrementalSettings, bool]] = None
        if args is not None:
            self._query = self._query.bindparams(**args)

    def one_or_none(self):
        self._finalizer = Finalizer.OneOrNone
        return self

    def one(self):
        self._finalizer = Finalizer.One
        return self

    def all(self, incremental: Union[bool, IncrementalSettings] = False):
        self._finalizer = Finalizer.All
        if isinstance(incremental, bool) and incremental:
            self._incremental_settings = IncrementalSettings(lookback_period=None)
        elif isinstance(incremental, IncrementalSettings):
            self._incremental_settings = incremental
        return self

    def execute_snowflake(self):
        # this import is safe because the only way we end up here is if we have a valid SnowflakeSource constructed,
        # which already gates this import
        import snowflake.connector

        from chalk.sql import SnowflakeSourceImpl

        source = typing.cast(SnowflakeSourceImpl, self._source)
        with snowflake.connector.connect(
            user=source.user,
            account=source.account_identifier,
            password=source.password,
            warehouse=source.warehouse,
            schema=source.schema,
            database=source.db,
        ) as con:
            return DataFrame(
                from_arrow(con.cursor().execute(self._original_query).fetch_arrow_all()).rename(
                    {k.upper(): v.root_fqn for k, v in self._fields.items()}
                )
            )

    def execute(self):
        from chalk.sql import SnowflakeSourceImpl

        if isinstance(self._source, SnowflakeSourceImpl) and (
            self._finalizer is None or self._finalizer == Finalizer.All
        ):
            return self.execute_snowflake()

        tuples = self._session.execute(self._query)
        assert isinstance(tuples, sqlalchemy.engine.CursorResult)
        cols = {k: Feature.from_root_fqn(v) if isinstance(v, str) else v for (k, v) in self._fields.items()}
        self._finalizer = self._finalizer or Finalizer.All

        if self._finalizer == Finalizer.All:
            data = {x.root_fqn: [] for x in cols.values()}
            for tup in tuples.all():
                for field in self._fields.keys():
                    data[cols[field].root_fqn].append(tup[field])
            df = DataFrame(data=data)
            return df

        tup = None
        if self._finalizer == Finalizer.One:
            tup = tuples.one()

        if self._finalizer == Finalizer.First:
            tup = tuples.first()

        if self._finalizer == Finalizer.OneOrNone:
            tup = tuples.one_or_none()

        return _construct_features(cols, tup, coerce_raw_values=True)


class ChalkQuery(ChalkQueryProtocol):
    _session: DBSessionProtocol
    _source: BaseSQLSourceProtocol
    _features: List[Feature]
    _finalizer: Optional[Finalizer]
    _incremental_settings: Optional[Union[IncrementalSettings, bool]]

    def __init__(
        self,
        features: List[Feature],
        session: DBSessionProtocol,
        source: BaseSQLSourceProtocol,
        raw_session: Optional[Any] = None,
    ):
        self._session = session
        self._raw_session = raw_session or session
        self._features = features
        self._finalizer = None
        self._source = source
        self._incremental_settings = None

    def first(self):
        self._session.update_query(lambda x: x.limit(1))
        self._finalizer = Finalizer.First
        return self

    def one_or_none(self):
        self._session.update_query(lambda x: x.limit(1))
        self._finalizer = Finalizer.OneOrNone
        return self

    def one(self):
        self._session.update_query(lambda x: x.limit(1))
        self._finalizer = Finalizer.One
        return self

    def all(self, incremental: Union[bool, IncrementalSettings] = False):
        self._finalizer = Finalizer.All
        if isinstance(incremental, bool) and incremental:
            self._incremental_settings = IncrementalSettings(lookback_period=None)
        elif isinstance(incremental, IncrementalSettings):
            self._incremental_settings = incremental
        return self

    def filter_by(self, **kwargs):
        self._session.update_query(lambda x: x.filter_by(**kwargs))
        return self

    def filter(self, *criterion):
        self._session.update_query(lambda x: x.filter(*criterion))
        return self

    def limit(self, *limits):
        self._session.update_query(lambda x: x.limit(*limits))
        return self

    def order_by(self, *clauses):
        self._session.update_query(lambda x: x.order_by(*clauses))
        return self

    def group_by(self, *clauses):
        self._session.update_query(lambda x: x.group_by(*clauses))
        return self

    def having(self, criterion):
        self._session.update_query(lambda x: x.having(*criterion))
        return self

    def union(self, *q):
        self._session.update_query(lambda x: x.union(*q))
        return self

    def union_all(self, *q):
        self._session.update_query(lambda x: x.union_all(*q))
        return self

    def intersect(self, *q):
        self._session.update_query(lambda x: x.intersect(*q))
        return self

    def intersect_all(self, *q):
        self._session.update_query(lambda x: x.intersect_all(*q))
        return self

    def join(self, target, *props, **kwargs):
        self._session.update_query(lambda x: x.join(target, *props, **kwargs))
        return self

    def outerjoin(self, target, *props, **kwargs):
        self._session.update_query(lambda x: x.outerjoin(target, *props, **kwargs))
        return self

    def select_from(self, *from_obj):
        self._session.update_query(lambda x: x.select_from(*from_obj))
        return self

    @staticmethod
    def _get_finalizer_fn(f: Optional[Finalizer]):
        if f == Finalizer.First:
            return lambda x: x.first()
        if f == Finalizer.All:
            return lambda x: x.all()
        if f == Finalizer.One:
            return lambda x: x.one()
        if f == Finalizer.OneOrNone:
            return lambda x: x.one_or_none()
        if f is None:
            return lambda x: x.all()
        raise ValueError(f"Unknown finalizer {f}")

    def execute(self):
        start = time.perf_counter()
        try:
            column_descriptions = self._session._session.column_descriptions
            self._session.update_query(self._get_finalizer_fn(self._finalizer))
            tuples = self._session.result()
            self._raw_session.close()
            fqn_to_feature = {f.fqn: f for f in self._features}
            cols = [fqn_to_feature[value["name"]] for value in column_descriptions if value["name"] in fqn_to_feature]

            if isinstance(tuples, list):
                lists = defaultdict(list)

                for tuple in tuples:
                    for col_i, x in enumerate(tuple):
                        if isinstance(x, Enum):
                            x = x.value
                        lists[col_i].append(x)
                return DataFrame({col: lists[i] for i, col in enumerate(cols)})

            return _construct_features(cols, tuples, coerce_raw_values=False)
        finally:
            _logger.debug(f"query.execute: {time.perf_counter() - start}")



from chalk.features import feature, features, unwrap_feature


@features
class CustomNameClass:
    id: str
    other: str = feature(name="other_name")


def test_custom_name():
    assert CustomNameClass(other="papaya").other == "papaya"
    assert CustomNameClass(id="4", other="papaya") == CustomNameClass(id="4", other="papaya")
    assert unwrap_feature(CustomNameClass.other).name == "other_name"
    assert str(CustomNameClass(other="papaya")) == "CustomNameClass(other=papaya)"
    assert str(CustomNameClass(id="4", other="papaya")) == "CustomNameClass(id=4, other=papaya)"
    assert dict(CustomNameClass(id="4", other="papaya")) == {
        "custom_name_class.id": "4",
        "custom_name_class.other_name": "papaya",
    }
    assert len(CustomNameClass(id="4", other="papaya")) == 2
    assert len(CustomNameClass(id="4")) == 1



from typing import Any, Callable, List, Optional, Type, Union

from pydantic import BaseModel

from chalk import MachineType
from chalk.features import Features


class StreamSource:
    message_model: Union[Type[Any], None] = None

    def config_to_json(self) -> Any:
        ...


class StreamResolver:
    registry: "List[StreamResolver]" = []

    def __init__(
        self,
        function_definition: str,
        fqn: str,
        filename: str,
        doc: Optional[str],
        source: StreamSource,
        fn: Callable,
        model: Union[BaseModel, None],
        output_features: Type[Features],
        environment: Optional[Union[List[str], str]],
        machine_type: Optional[MachineType],
    ):
        super(StreamResolver, self).__init__()
        self.function_definition = function_definition
        self.fqn = fqn
        self.filename = filename
        self.source = source
        self.fn = fn
        self.model = model
        self.environment = environment
        self.doc = doc
        self.machine_type = machine_type
        self.output_features = output_features

    def __eq__(self, other):
        return isinstance(other, StreamResolver) and self.fqn == other.fqn

    def __hash__(self):
        return hash(self.fqn)

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    def __repr__(self):
        return f"StreamResolver(name={self.fqn})"



from datetime import datetime

from chalk.features.feature_field import Feature

__all__ = [
    "CHALK_TS_FEATURE",
    "ID_FEATURE",
    "OBSERVED_AT_FEATURE",
    "REPLACED_OBSERVED_AT_FEATURE",
    "PSEUDOFEATURES",
]

CHALK_TS_FEATURE = Feature(
    name="CHALK_TS", namespace="__chalk__", typ=datetime, max_staleness=None, etl_offline_to_online=False
)
ID_FEATURE = Feature(name="__id__", namespace="__chalk__", typ=str, max_staleness=None, etl_offline_to_online=False)
OBSERVED_AT_FEATURE = Feature(
    name="__observed_at__", namespace="__chalk__", typ=datetime, max_staleness=None, etl_offline_to_online=False
)
REPLACED_OBSERVED_AT_FEATURE = Feature(
    name="__replaced_observed_at__",
    namespace="__chalk__",
    typ=datetime,
    max_staleness=None,
    etl_offline_to_online=False,
)

PSEUDOFEATURES = [CHALK_TS_FEATURE, ID_FEATURE, OBSERVED_AT_FEATURE, REPLACED_OBSERVED_AT_FEATURE]



from pydantic import BaseModel

from chalk.features import Features, features
from chalk.streams import KafkaSource, stream


@features
class StreamFeatures:
    scalar_feature: str


class KafkaMessage(BaseModel):
    val_a: str


s = KafkaSource(broker=[], topic=[], message_model=KafkaMessage)


@stream(source=s)
def fn(message: KafkaMessage) -> Features[StreamFeatures.scalar_feature]:
    return StreamFeatures(
        scalar_feature=message.val_a,
    )


def test_callable():
    assert fn(KafkaMessage(val_a="hello")) == StreamFeatures(scalar_feature="hello")


def test_parsed_source():
    assert fn.source == s



from typing import Any, Dict, Optional, Union

from sqlalchemy.engine.url import URL

from chalk.client.client_impl import ChalkConfigurationException
from chalk.integrations.named import load_integration_variable
from chalk.sql.base.sql_source import BaseSQLSource, TableIngestMixIn


class MySQLSourceImpl(BaseSQLSource, TableIngestMixIn):
    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[str] = None,
        db: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        name: Optional[str] = None,
    ):
        try:
            import pymysql
        except ModuleNotFoundError:
            raise ChalkConfigurationException.missing_dependency("chalkpy[mysql]")
        self.host = host or load_integration_variable(name="MYSQL_HOST", integration_name=name)
        self.port = port or load_integration_variable(name="MYSQL_TCP_PORT", integration_name=name)
        self.db = db or load_integration_variable(name="MYSQL_DATABASE", integration_name=name)
        self.user = user or load_integration_variable(name="MYSQL_USER", integration_name=name)
        self.password = password or load_integration_variable(name="MYSQL_PWD", integration_name=name)
        self.ingested_tables: Dict[str, Any] = {}
        super(MySQLSourceImpl, self).__init__()

    def local_engine_url(self) -> Union[str, URL]:
        return URL.create(
            drivername="mysql+pymysql",
            username=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.db,
        )



import ast
import functools
import inspect
import re
import sys
import textwrap
import threading
import types
from datetime import datetime
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from chalk.features.feature_field import Feature
from chalk.features.feature_set import Features, FeatureSetBase
from chalk.features.feature_wrapper import FeatureWrapper
from chalk.features.parsed_annotation import FeatureTypeAnnotation
from chalk.features.tag import Tags
from chalk.utils.collections import ensure_tuple
from chalk.utils.duration import Duration
from chalk.utils.metaprogramming import MISSING, create_fn, field_assign, set_new_attribute
from chalk.utils.string import removeprefix, to_snake_case

T = TypeVar("T")

__all__ = ["features"]


@overload
def features(
    *,
    owner: Optional[str] = None,
    tags: Optional[Tags] = None,
    etl_offline_to_online: bool = ...,
    max_staleness: Optional[Duration] = ...,
) -> Callable[[Type[T]], Type[T]]:
    ...


@overload
def features(cls: Type[T]) -> Type[T]:
    ...


def features(
    cls: Optional[Type[T]] = None,
    *,
    owner: Optional[str] = None,
    tags: Optional[Tags] = None,
    etl_offline_to_online: Optional[bool] = None,
    max_staleness: Optional[Duration] = ...,
) -> Union[Callable[[Type[T]], Type[T]], Type[T]]:
    """Returns the same class as was passed in, with dunder methods
    added based on the fields defined in the class.

    Examines PEP 526 __annotations__ to determine fields.
    """

    def wrap(c: Type[T]) -> Type[T]:
        updated_class = _process_class(
            cls=c,
            owner=owner,
            tags=ensure_tuple(tags),
            etl_offline_to_online=etl_offline_to_online,
            max_staleness=max_staleness,
        )
        set_new_attribute(
            cls=updated_class,
            name="namespace",
            value=to_snake_case(updated_class.__name__),
        )
        assert issubclass(updated_class, Features)
        FeatureSetBase.registry[updated_class.namespace] = updated_class

        primary_feature = _discover_feature(
            updated_class,
            "primary",
            lambda f: f.primary,
            lambda f: f.name == "id" and not f.has_resolved_join and not f.is_feature_time,
        )
        if primary_feature is not None:
            primary_feature.primary = True

        timestamp_feature = _discover_feature(
            updated_class,
            "feature time",
            lambda f: f.is_feature_time,
            lambda f: f.name == "ts" and not f.has_resolved_join,
        )
        set_new_attribute(cls=updated_class, name="__chalk_primary__", value=primary_feature)
        set_new_attribute(cls=updated_class, name="__chalk_ts__", value=timestamp_feature)

        return cast(Type[T], updated_class)

    # See if we're being called as @features or @features().
    if cls is None:
        # We're called with parens.
        return wrap

    # We're called as @features without parens.
    return wrap(cls)


def _discover_feature(cls: Type[Features], name: str, *conditions: Callable[[Feature], bool]):
    for cond in conditions:
        features = [c for c in cls.features if cond(c)]
        if len(features) == 1:
            return features[0]
        if len(features) > 1:
            raise ValueError(
                f"Multiple {name} features are not supported in {cls.__name__}: "
                + ", ".join(f"{cls.__name__}.{x.name}" for x in features)
            )
    return None


def _init_param(f: Feature):
    # Return the __init__ parameter string for this field.  For
    # example, the equivalent of 'x:int=3' (except instead of 'int',
    # reference a variable set to int, and instead of '3', reference a
    # variable set to 3).
    return f"{f.attribute_name}:_type_{f.attribute_name}=MISSING"


def _getattribute_fn(_globals: Dict[str, Any]):
    # If calling getattr() on an instance for a feature name, do NOT return a class-level FeatureWrapper
    # Instead, raise an attribute error
    return create_fn(
        name="__getattribute__",
        args=["self", "attribute_name"],
        body=[
            "o = object.__getattribute__(self, attribute_name)",
            "if isinstance(o, FeatureWrapper):",
            "    raise AttributeError(f'Feature {attribute_name} is not defined on this instance of class {type(self).__name__}')",
            "return o",
        ],
        _globals=_globals,
        _locals={
            "FeatureWrapper": FeatureWrapper,
        },
        return_type=Any,
    )


def _init_fn(fields: List[Feature], self_name: str, _globals: Dict[str, Any]):
    # fields contains both real fields and InitVar pseudo-fields.

    # Make sure we don't have fields without defaults following fields
    # with defaults.  This actually would be caught when exec-ing the
    # function source code, but catching it here gives a better error
    # message, and future-proofs us in case we build up the function
    # using ast.

    _locals: MutableMapping[str, Any] = {f"_type_{f.attribute_name}": f.typ and f.typ.annotation for f in fields}
    _locals.update(
        {
            "MISSING": MISSING,
            # '_HAS_DEFAULT_FACTORY': _HAS_DEFAULT_FACTORY,
        }
    )

    body_lines = []
    for f in fields:
        assert f.attribute_name is not None
        line = field_assign(f.attribute_name, f.attribute_name, self_name)
        # line is None means that this field doesn't require
        # initialization (it's a pseudo-field).  Just skip it.
        if line:
            body_lines.append(line)

    # If no body lines, use 'pass'.
    if not body_lines:
        body_lines = ["pass"]

    # Add the keyword-only args.  Because the * can only be added if
    # there's at least one keyword-only arg, there needs to be a test here
    # (instead of just concatenting the lists together).
    _init_params = [_init_param(f) for f in fields]
    return create_fn(
        name="__init__",
        args=[self_name] + _init_params,
        body=body_lines,
        _locals=_locals,
        _globals=_globals,
        return_type=None,
    )


def _parse_tags(ts: str) -> List[str]:
    ts = re.sub(",", " ", ts)
    ts = re.sub(" +", " ", ts.strip())
    return [xx.strip() for xx in ts.split(" ")]


def _get_field(
    cls: Type,
    annotation_name: str,
    comments: Mapping[str, str],
    class_owner: Optional[str],
    class_tags: Optional[Tuple[str, ...]],
    class_etl_offline_to_online: Optional[bool],
    class_max_staleness: Optional[Duration],
):
    # Return a Field object for this field name and type.  ClassVars and
    # InitVars are also returned, but marked as such (see f._field_type).
    # default_kw_only is the value of kw_only to use if there isn't a field()
    # that defines it.

    # If the default value isn't derived from Field, then it's only a
    # normal default value.  Convert it to a Field().
    default = getattr(cls, annotation_name, MISSING)

    if isinstance(default, Feature):
        f = default
    else:
        if isinstance(default, types.MemberDescriptorType):
            # This is a field in __slots__, so it has no default value.
            default = MISSING
        f = Feature(name=annotation_name, namespace=to_snake_case(cls.__name__))

    # Only at this point do we know the name and the type.  Set them.
    f.namespace = to_snake_case(cls.__name__)
    f.typ = FeatureTypeAnnotation(cls, annotation_name)
    f.features_cls = cls
    f.attribute_name = annotation_name
    comment_for_feature = comments.get(annotation_name)
    comment_based_description = None
    comment_based_owner = None
    comment_based_tags = None

    if comment_for_feature is not None:
        comment_lines = []
        for line in comment_for_feature.splitlines():
            stripped_line = line.strip()
            if stripped_line.startswith(":owner:"):
                comment_based_owner = removeprefix(stripped_line, ":owner:").strip()
            elif stripped_line.startswith(":tags:"):
                parsed = _parse_tags(removeprefix(stripped_line, ":tags:"))
                if len(parsed) > 0:
                    if comment_based_tags is None:
                        comment_based_tags = parsed
                    else:
                        comment_based_tags.extend(parsed)
            else:
                comment_lines.append(line)

        comment_based_description = "\n".join(comment_lines)

    if f.description is None and comment_based_description is not None:
        f.description = comment_based_description

    if comment_based_tags is not None:
        if f.tags is None:
            f.tags = comment_based_tags
        else:
            f.tags.extend(comment_based_tags)

    if class_tags is not None:
        if f.tags is None:
            f.tags = list(class_tags)
        else:
            f.tags.extend(class_tags)

    if f.owner is not None and comment_based_owner is not None:
        raise ValueError(
            (
                f"Owner for {f.namespace}.{f.name} specified both on the feature and in the comment. "
                f"Please use only one of these two."
            )
        )
    elif f.owner is None:
        f.owner = comment_based_owner or class_owner

    if f.max_staleness is Ellipsis:
        f.max_staleness = class_max_staleness if class_max_staleness is not Ellipsis else None

    if f.etl_offline_to_online is None:
        f.etl_offline_to_online = False if class_etl_offline_to_online is None else class_etl_offline_to_online
    return f


def _recursive_repr(user_function: Callable[[T], str]) -> Callable[[T], str]:
    # Decorator to make a repr function return "..." for a recursive
    # call.
    repr_running = set()

    @functools.wraps(user_function)
    def wrapper(self: T):
        key = id(self), threading.get_ident()
        if key in repr_running:
            return "..."
        repr_running.add(key)
        try:
            result = user_function(self)
        finally:
            repr_running.discard(key)
        return result

    return wrapper


def _repr_fn(fields: List[Feature], _globals: Dict[str, Any]):
    tuples = ",".join(f"('{f.attribute_name}', getattr(self, '{f.attribute_name}', MISSING))" for f in fields)
    fn = create_fn(
        name="__repr__",
        args=["self"],
        body=[
            'return f"{self.__class__.__qualname__}(" + '
            + f"', '.join(f'{{x[0]}}={{x[1]}}' for x in [{tuples}] if x[1] != MISSING)"
            + '+ ")"'
        ],
        _globals=_globals,
        _locals={"MISSING": MISSING},
    )
    return _recursive_repr(fn)


def _eq_fn(fields: List[Feature], _globals: Dict[str, Any]):
    cmp_str = " and ".join(
        f"(getattr(self, '{f.attribute_name}', MISSING) == getattr(other, '{f.attribute_name}', MISSING))"
        for f in fields
    )
    if not cmp_str:
        # if there are no features in this class
        cmp_str = "True"
    return create_fn(
        name="__eq__",
        args=["self", "other"],
        body=["if not isinstance(other, type(self)):", "    return NotImplemented", f"return {cmp_str}"],
        _globals=_globals,
        _locals={
            "MISSING": MISSING,
        },
    )


def _len_fn(_globals: Dict[str, Any]):
    return create_fn(
        name="__len__",
        args=["self"],
        body=[
            "count = 0",
            f"for f in self.features:",
            "    if hasattr(self, f.attribute_name):",
            "        count += 1",
            "return count",
        ],
        _globals=_globals,
        _locals={},
    )


def _iter_fn(fields: List[Feature], _globals: Mapping[str, Any]):
    return create_fn(
        "__iter__",
        args=["self"],
        body=[
            f"for __chalk_f__ in self.features:",
            f"    if hasattr(self, __chalk_f__.attribute_name) and type(__chalk_f__).__name__ == 'Feature' and not __chalk_f__.is_has_one and not __chalk_f__.is_has_many:",
            f"        yield __chalk_f__.fqn, getattr(self, __chalk_f__.attribute_name)",
        ],
    )


def _parse_annotation_comments(cls: Type) -> Mapping[str, str]:
    source = textwrap.dedent(inspect.getsource(cls))

    source_lines = source.splitlines()
    tree = ast.parse(source)
    if len(tree.body) != 1:
        return {}

    comments_for_annotations: Dict[str, str] = {}
    class_def = tree.body[0]
    if isinstance(class_def, ast.ClassDef):
        for stmt in class_def.body:
            if isinstance(stmt, ast.AnnAssign) and isinstance(stmt.target, ast.Name):
                line = stmt.lineno - 2
                comments: List[str] = []
                while line >= 0 and source_lines[line].strip().startswith("#"):
                    comment = source_lines[line].strip().strip("#").strip()
                    comments.insert(0, comment)
                    line -= 1

                if len(comments) > 0:
                    comments_for_annotations[stmt.target.id] = textwrap.dedent("\n".join(comments))

    return comments_for_annotations


def _process_class(
    cls: Type[T],
    owner: Optional[str],
    tags: Tuple[str, ...],
    etl_offline_to_online: Optional[bool],
    max_staleness: Optional[Duration],
) -> Type[T]:
    cls_annotations = cls.__dict__.get("__annotations__", {})

    if cls.__module__ in sys.modules:
        _globals = sys.modules[cls.__module__].__dict__
    else:
        _globals = {}

    # Find feature times that weren't annotated.
    for (name, member) in inspect.getmembers(cls):
        if name not in cls_annotations and isinstance(member, Feature):
            # All feature types need annotations, except for datetimes, which we can automatically infer
            if member.typ is not None and member.typ.is_parsed and issubclass(member.typ.underlying, datetime):
                cls_annotations[name] = datetime
            else:
                raise TypeError(f"Member {name} is missing an annotation")

    # If we pass this function something that isn't a class, it could raise
    try:
        comments = _parse_annotation_comments(cls)
    except:
        comments = {}

    cls_fields = [
        _get_field(
            cls=cls,
            annotation_name=name,
            comments=comments,
            class_owner=owner,
            class_tags=tags,
            class_etl_offline_to_online=etl_offline_to_online,
            class_max_staleness=max_staleness,
        )
        for name in cls_annotations
    ]

    set_new_attribute(
        cls=cls,
        name="__init__",
        value=_init_fn(
            fields=cls_fields,
            # The name to use for the "self"
            # param in __init__.  Use "self"
            # if possible.
            self_name="self",
            _globals=_globals,
        ),
    )

    set_new_attribute(cls=cls, name="__repr__", value=_repr_fn(fields=cls_fields, _globals=_globals))
    set_new_attribute(cls=cls, name="__eq__", value=_eq_fn(fields=cls_fields, _globals=_globals))
    set_new_attribute(cls=cls, name="__hash__", value=None)
    set_new_attribute(cls=cls, name="__iter__", value=_iter_fn(fields=cls_fields, _globals=_globals))

    namespace = to_snake_case(cls.__name__)
    set_new_attribute(cls=cls, name="__chalk_namespace__", value=namespace)
    set_new_attribute(cls=cls, name="__chalk_owner__", value=owner)
    set_new_attribute(cls=cls, name="__chalk_tags__", value=list(tags))
    set_new_attribute(
        cls=cls, name="__chalk_max_staleness__", value=max_staleness if max_staleness is not Ellipsis else None
    )
    set_new_attribute(
        cls=cls,
        name="__chalk_etl_offline_to_online__",
        value=etl_offline_to_online if etl_offline_to_online is not Ellipsis else False,
    )
    set_new_attribute(cls=cls, name="features", value=cls_fields)
    set_new_attribute(cls=cls, name="__is_features__", value=True)
    set_new_attribute(cls=cls, name="__len__", value=_len_fn(_globals=_globals))
    set_new_attribute(cls=cls, name="__getattribute__", value=_getattribute_fn(_globals=_globals))

    for f in cls_fields:
        assert f.attribute_name is not None
        # Wrap all class features with FeatureWrapper
        setattr(cls, f.attribute_name, FeatureWrapper(f))

    return cls



import concurrent.futures
import logging
import unittest
import unittest.mock

from chalk.utils.log_with_context import Logger, add_logging_context

LOGGER = logging.getLogger()

LOG_METHOD_NAMES = [
    "debug",
    "info",
    "warning",
    "warn",
    "error",
    "critical",
    "exception",
]


def process_worker(val):
    """See how our logger interacts with multiprocessing."""
    base_logger = unittest.mock.Mock(spec=LOGGER)
    logger = Logger(logger=base_logger)

    logger.debug("1")
    base_logger.debug.assert_called_with("1", extra={}, stacklevel=3)
    with add_logging_context(val=val):
        logger.debug("2")
        base_logger.debug.assert_called_with("2", extra=dict(val=val), stacklevel=3)


class TestLogger(unittest.TestCase):
    """Unit tests for :py:class:`log_with_context.Logger`."""

    def setUp(self):
        self.base_logger = unittest.mock.Mock(spec=LOGGER)
        self.logger = Logger(logger=self.base_logger)

    def gl(self, method_name: str):
        """Execute a given method on our test logger."""
        return getattr(self.logger, method_name)

    def gb(self, method_name: str):
        """Execute a given method on our base logger mock."""
        return getattr(self.base_logger, method_name)

    def test_add_logging_context(self):
        """Test that :py:class:`add_logging_context` works."""
        for method_name in LOG_METHOD_NAMES:

            with self.subTest(name=f"logger.{method_name}"):
                self.gl(method_name)("1")
                self.gb(method_name).assert_called_with("1", extra={}, stacklevel=3)

                with add_logging_context(a=1):
                    self.gl(method_name)("2")
                    self.gb(method_name).assert_called_with("2", extra=dict(a=1), stacklevel=3)

                    with add_logging_context(b=2):
                        self.gl(method_name)("3")
                        self.gb(method_name).assert_called_with("3", extra=dict(a=1, b=2), stacklevel=3)

                        with add_logging_context(a=4, c=3):
                            self.gl(method_name)("4")
                            self.gb(method_name).assert_called_with("4", extra=dict(a=4, b=2, c=3), stacklevel=3)

                        self.gl(method_name)("5")
                        self.gb(method_name).assert_called_with("5", extra=dict(a=1, b=2), stacklevel=3)

                    self.gl(method_name)("6")
                    self.gb(method_name).assert_called_with("6", extra=dict(a=1), stacklevel=3)

                self.gl(method_name)("7")
                self.gb(method_name).assert_called_with("7", extra={}, stacklevel=3)

            self.gl(method_name)("8")
            self.gb(method_name).assert_called_with("8", extra={}, stacklevel=3)

    def test_inline_extra(self):
        """Test that we can add one-off additions to our context."""
        for method_name in LOG_METHOD_NAMES:
            with self.subTest(name=f"logger.{method_name}"):
                self.gl(method_name)("%s", method_name, extra=dict(hello=method_name))
                self.gb(method_name).assert_called_with("%s", method_name, extra=dict(hello=method_name), stacklevel=3)

    def test_thread_local(self):
        """Test that our logging context is indeed thread-local."""

        def thread_worker(val):
            self.logger.debug("1")
            self.base_logger.debug.assert_called_with("1", extra={}, stacklevel=3)
            with add_logging_context(val=val):
                self.logger.debug("2")
                self.base_logger.debug.assert_called_with("2", extra=dict(val=val), stacklevel=3)

        with concurrent.futures.ThreadPoolExecutor(max_workers=2) as exc:
            with add_logging_context(a=1):
                list(exc.map(thread_worker, [1, 2]))

    def test_process_local(self):
        """Test how our logger works with multiple processes."""
        with concurrent.futures.ProcessPoolExecutor(max_workers=2) as exc:
            with add_logging_context(a=1):
                list(exc.map(process_worker, [1, 2]))



from datetime import datetime

import pytest

from chalk import description, is_primary, owner, tags
from chalk.features import Features, feature, feature_time, features, unwrap_feature


@features(tags="t", owner="andy@chalk.ai")
class WowFS:
    # This is a really neat description of something
    something: str

    # This is also really neat and cool
    something_else: str
    nocomment: str

    nope: str

    assigned: str = feature(tags="a")

    # bizarre
    bizarre: str

    # goofy
    goofy: str = feature(owner="yo")

    # now with feature
    assigned_comment: str = feature(tags=["3", "4"])

    # implicit comment
    explicit: str = feature(description="explicit comment")

    # Multiline
    # Neat and, verily, cool
    #
    # Hello
    assigned_comment_multiline: str = feature()

    # now with feature time
    time: datetime = feature_time()

    nope_nope: datetime  # Datetime field that is not feature time


@features(owner="elliot@chalk.ai")
class OwnerFeatures:
    plain: str
    cached: str = feature(max_staleness="3d")
    andy: str = feature(owner="andy@chalk.ai")

    ft: datetime = feature_time()


@features(tags=["1", "2"])
class TagFeatures:
    empty: str
    one: str = feature(tags="one")
    many: str = feature(tags=["a", "b"])

    ft: datetime = feature_time()


@features
class CommentBaseOwner:
    id: str
    # I'm a cool comment!
    # :owner: elliot@chalk.ai
    empty: str

    # I'm a cool comment!
    # :tags: pii group:risk
    email: str

    # :tags: pii, group:risk
    email_commas: str
    # :tags: pii
    email_single: str

    # :tags: pii
    email_all_kinds: str = feature(tags=["hello"])


@features(max_staleness="4d")
class MaxStalenessFeatures:
    id: int
    name: str = feature(max_staleness=None)
    woohoo: str = feature(max_staleness="30d")


@features(etl_offline_to_online=True)
class ETLFeatures:
    id: int
    name: str = feature(etl_offline_to_online=False)
    woohoo: str = feature(etl_offline_to_online=True)


def test_class_etl_offline_to_online():
    assert unwrap_feature(ETLFeatures.id).etl_offline_to_online is True
    assert unwrap_feature(ETLFeatures.name).etl_offline_to_online is False
    assert unwrap_feature(ETLFeatures.woohoo).etl_offline_to_online is True
    assert issubclass(ETLFeatures, Features)
    assert ETLFeatures.__chalk_etl_offline_to_online__ is True
    assert issubclass(MaxStalenessFeatures, Features)
    assert MaxStalenessFeatures.__chalk_etl_offline_to_online__ is None


def test_class_max_staleness():
    assert unwrap_feature(MaxStalenessFeatures.id).max_staleness == "4d"
    assert unwrap_feature(MaxStalenessFeatures.name).max_staleness is None
    assert unwrap_feature(MaxStalenessFeatures.woohoo).max_staleness == "30d"
    assert issubclass(MaxStalenessFeatures, Features)
    assert MaxStalenessFeatures.__chalk_max_staleness__ == "4d"
    assert issubclass(ETLFeatures, Features)
    assert ETLFeatures.__chalk_max_staleness__ is None


def test_primary():
    assert is_primary(CommentBaseOwner.id)
    assert not is_primary(CommentBaseOwner.email)


def test_comment_based_owner():
    assert "elliot@chalk.ai" == unwrap_feature(CommentBaseOwner.empty).owner == owner(CommentBaseOwner.empty)
    assert ["pii", "group:risk"] == unwrap_feature(CommentBaseOwner.email).tags == tags(CommentBaseOwner.email)
    assert ["pii"] == unwrap_feature(CommentBaseOwner.email_single).tags == tags(CommentBaseOwner.email_single)
    assert (
        ["hello", "pii"]
        == unwrap_feature(CommentBaseOwner.email_all_kinds).tags
        == tags(CommentBaseOwner.email_all_kinds)
    )
    with pytest.raises(ValueError):

        @features
        class BadFeatureClass:
            # :owner: elliot@chalk.ai
            doubly_owned: str = feature(owner="d")


def test_parse_descriptions():
    assert "bizarre" == unwrap_feature(WowFS.bizarre).description == description(WowFS.bizarre)
    assert (
        "This is a really neat description of something"
        == unwrap_feature(WowFS.something).description
        == description(WowFS.something)
    )
    assert "explicit comment" == unwrap_feature(WowFS.explicit).description == description(WowFS.explicit)
    assert (
        "This is also really neat and cool"
        == unwrap_feature(WowFS.something_else).description
        == description(WowFS.something_else)
    )
    assert (
        "now with feature" == unwrap_feature(WowFS.assigned_comment).description == description(WowFS.assigned_comment)
    )
    assert "goofy" == unwrap_feature(WowFS.goofy).description == description(WowFS.goofy)
    assert (
        """Multiline
Neat and, verily, cool

Hello"""
        == unwrap_feature(WowFS.assigned_comment_multiline).description
        == description(WowFS.assigned_comment_multiline)
    )


def test_class_owner():
    assert "elliot@chalk.ai" == OwnerFeatures.__chalk_owner__ == owner(OwnerFeatures)
    assert "elliot@chalk.ai" == unwrap_feature(OwnerFeatures.plain).owner == owner(OwnerFeatures.plain)
    assert "elliot@chalk.ai" == unwrap_feature(OwnerFeatures.cached).owner == owner(OwnerFeatures.cached)
    assert "andy@chalk.ai" == unwrap_feature(OwnerFeatures.andy).owner == owner(OwnerFeatures.andy)


def test_class_tags():
    assert [] == tags(OwnerFeatures) == OwnerFeatures.__chalk_tags__
    assert ["1", "2"] == TagFeatures.__chalk_tags__ == tags(TagFeatures)
    assert ["1", "2"] == unwrap_feature(TagFeatures.empty).tags == tags(TagFeatures.empty)
    assert ["one", "1", "2"] == unwrap_feature(TagFeatures.one).tags == tags(TagFeatures.one)
    assert ["a", "b", "1", "2"] == unwrap_feature(TagFeatures.many).tags == tags(TagFeatures.many)
    assert ["a", "t"] == unwrap_feature(WowFS.assigned).tags == tags(WowFS.assigned)
    assert ["t"] == unwrap_feature(WowFS.nope).tags == tags(WowFS.nope)
    assert ["3", "4", "t"] == unwrap_feature(WowFS.assigned_comment).tags == tags(WowFS.assigned_comment)



from chalk.streams._file_source import FileSource
from chalk.streams._kafka_source import KafkaSource
from chalk.streams._stream import stream
from chalk.streams._types import StreamResolver, StreamSource
from chalk.streams._windows import Windowed, windowed

__all__ = [
    "FileSource",
    "KafkaSource",
    "stream",
    "StreamResolver",
    "StreamSource",
    "Windowed",
    "windowed",
]



import importlib
import importlib.util
import logging
import os
import sys
import traceback
from pathlib import Path
from typing import List

from pydantic import BaseModel

from chalk.gitignore.gitignore_parser import parse_gitignore
from chalk.utils.paths import get_directory_root

_logger = logging.getLogger(__name__)


def py_path_to_module(path: Path, repo_root: Path) -> str:
    try:
        p = path.relative_to(repo_root)
    except ValueError:
        p = path
    return str(p)[: -len(".py")].replace("./", "").replace("/", ".")


class FailedImport(BaseModel):
    filename: str
    module: str
    traceback: str


def import_all_python_files() -> List[FailedImport]:
    project_root = get_directory_root()
    return import_all_python_files_from_dir(project_root=project_root)


def import_all_python_files_from_dir(project_root: Path) -> List[FailedImport]:
    gitignore_path = project_root / ".gitignore"
    chalk_ignore_path = project_root / ".chalkignore"
    matching_functions = []
    if gitignore_path.exists():
        matching_functions.append(parse_gitignore(str(gitignore_path)))

    if chalk_ignore_path.exists():
        matching_functions.append(parse_gitignore(str(chalk_ignore_path)))

    is_relevant = lambda x: all((not match(x) for match in matching_functions))

    cwd = os.getcwd()
    os.chdir(project_root)
    repo_root = Path(project_root)

    # If we don't import both of these, we get in trouble.
    sys.path.append(str(repo_root.resolve()))
    sys.path.append(str(repo_root.parent.resolve()))

    repo_files = sorted({p.resolve() for p in repo_root.glob("**/*.py") if p.is_file()})

    _logger.debug(f"REPO_ROOT: {repo_root.resolve()}")
    # _logger.debug(f"REPO_FILES: {repo_files}")

    venv = os.environ.get("VIRTUAL_ENV")
    module_paths = [
        (py_path_to_module(repo_file, repo_root), repo_file)
        for repo_file in repo_files
        if venv is None or Path(venv) not in repo_file.parents
    ]

    errors: List[FailedImport] = []
    for (module_path, filename) in module_paths:
        if module_path.startswith(".eggs") or module_path.startswith("venv") or not is_relevant(filename):
            continue

        try:
            importlib.import_module(module_path)
        except Exception:
            filename = filename.resolve()
            # relevant_file = any("from chalk." or "import chalk." in c for c in f)
            # if relevant_file:
            ex_type, ex_value, ex_traceback = sys.exc_info()
            tb = traceback.extract_tb(ex_traceback)
            line = 0
            for i, l in enumerate(tb):
                if filename == Path(l.filename).resolve():
                    line = i
                    break

            relevant_traceback = f"""Exception in module {module_path}:
{os.linesep.join(traceback.format_tb(ex_traceback)[line:])}
\n{ex_type and ex_type.__name__}: {str(ex_value)}
"""
            errors.append(
                FailedImport(
                    traceback=relevant_traceback,
                    filename=str(filename),
                    module=module_path,
                )
            )

            _logger.debug(f"Failed while importing {module_path}", exc_info=True)
            continue

    os.chdir(cwd)
    return errors



from typing import Any, Callable, MutableMapping

from sqlalchemy import create_engine
from sqlalchemy.orm import Session, sessionmaker

from chalk.sql.base.protocols import BaseSQLSourceProtocol, DBSessionMakerProtocol, DBSessionProtocol


class DBSession(DBSessionProtocol):
    def __init__(self, s: Session):
        self._session = s
        self._raw_session = s

    def update_query(self, f: Callable[[Session], Session]) -> None:
        self._session = f(self._session)

    def result(self) -> Any:
        return self._session

    def execute(self, q: Any) -> Any:
        return self._session.execute(q)

    def close(self):
        return self._raw_session.close()


class DBSessionMaker(DBSessionMakerProtocol):
    def __init__(self):
        self._session_makers: MutableMapping[BaseSQLSourceProtocol, sessionmaker] = {}

    def get_session(self, source: BaseSQLSourceProtocol) -> DBSessionProtocol:
        if source not in self._session_makers:
            engine = create_engine(
                source.local_engine_url(),
                echo=False,
                **source.engine_args(),
            )
            sm: sessionmaker = sessionmaker(bind=engine)
            self._session_makers[source] = sm

        sm = self._session_makers[source]
        return DBSession(sm())



from typing import Generic, Type, TypeVar

T = TypeVar("T")


class WindowedMeta(Generic[T], type):
    kind: Type[T]

    def __getitem__(cls, item: Type[T]) -> "WindowedMeta[T]":
        cls.kind = item
        return cls


class Windowed(Generic[T], metaclass=WindowedMeta[T]):
    kind: Type[T]

    def __getitem__(self, item) -> "Windowed[T]":
        pass

    def __init__(self, *buckets: str):
        self.buckets = buckets

    def __call__(self, selected: str) -> Type[T]:
        return SelectedWindow(kind=self, selected=selected)


class SelectedWindow:
    def __init__(self, kind: Windowed, selected: str):
        self.kind = kind
        self.selected = selected


def windowed(*buckets: str) -> Windowed[T]:
    return Windowed(*buckets)



from __future__ import annotations

import collections.abc
import dataclasses
import enum
import logging
from datetime import date, datetime, timezone
from typing import TYPE_CHECKING, Any, Type, Union, cast, get_origin, get_type_hints

import cattrs
import pendulum
import polars as pl
import pydantic
import pytz
from dateutil import parser
from pendulum.tz.timezone import Timezone
from polars.internals.construction import is_namedtuple
from sqlalchemy import Boolean, Float, Integer, Text
from sqlalchemy.types import TypeEngine

from chalk.utils.enum import get_enum_value_type
from chalk.utils.sqlalchemy import UtcDateTime

if TYPE_CHECKING:
    from chalk.features import Feature

try:
    import attrs
except ImportError:
    # Imports not available. Attrs is not required.
    attrs = None

try:
    import pandas as pd
except ImportError:
    pd = None

try:
    import numpy as np
except ImportError:
    np = None

_log = logging.getLogger(__name__)


class FeatureCodec:
    def __init__(self):
        self.converter = cattrs.Converter()
        self.converter.register_structure_hook(datetime, lambda v, _: parser.isoparse(v))
        self.converter.register_unstructure_hook(
            datetime,
            lambda v: (
                cast(datetime, v) if cast(datetime, v).tzinfo else pytz.utc.localize(cast(datetime, v))
            ).isoformat(),
        )

    def _default_encode(self, value: Any):
        if isinstance(value, set):
            return {self._default_encode(x) for x in value}
        if isinstance(value, list):
            return [self._default_encode(x) for x in value]
        if isinstance(value, (str, int, float)):
            return value
        if isinstance(value, enum.Enum):
            return self._default_encode(value.value)
        if isinstance(value, datetime):
            tz = value.tzinfo or (datetime.now(timezone.utc).astimezone().tzinfo)
            assert tz is not None
            return pendulum.instance(value, cast(Timezone, tz)).isoformat()
        if isinstance(value, date):
            return value.isoformat()
        if np is not None and isinstance(value, np.integer):
            return int(value)
        if np is not None and isinstance(value, np.floating):
            return float(value)
        if pd is not None and isinstance(value, pd.Timestamp):
            return pendulum.instance(value.to_pydatetime()).isoformat()
        if isinstance(value, pydantic.BaseModel):
            return value.dict()
        if (attrs is not None and attrs.has(type(value))) or dataclasses.is_dataclass(value):
            return self.converter.unstructure(value)
        if isinstance(value, collections.abc.Iterable):
            return [self._default_encode(x) for x in value]
        raise TypeError(f"Unable to encode value of type {type(value).__name__}")

    def encode(
        self,
        feature: Feature,
        value: Any,
    ):
        if value is None:
            return None

        if feature.encoder is not None:
            return feature.encoder(value)

        return self._default_encode(value)

    def encode_fqn(self, fqn: str, value: Any):
        from chalk.features import Feature

        return self.encode(Feature.from_root_fqn(fqn), value)

    def _default_decode_value(
        self,
        feature: Feature,
        value: Any,
    ):
        assert feature.typ is not None
        if issubclass(feature.typ.underlying, enum.Enum):
            value = feature.typ.underlying(value)
        if issubclass(feature.typ.underlying, datetime):
            value = parser.isoparse(value)
        elif issubclass(feature.typ.underlying, date):
            # note: datetime is a subclass of date, so we must be careful to decode accordingly
            value = date.fromisoformat(value)
        if issubclass(feature.typ.underlying, pydantic.BaseModel):
            return feature.typ.underlying(**value)
        elif (attrs is not None and attrs.has(feature.typ.underlying)) or dataclasses.is_dataclass(
            feature.typ.underlying
        ):
            return self.converter.structure(value, feature.typ.underlying)
        if not isinstance(value, feature.typ.underlying):
            raise TypeError(f"Unable to decode value {value} to type {feature.typ.underlying.__name__}")
        return value

    def _default_decode(
        self,
        feature: Feature,
        value: Any,
    ):
        assert feature.typ is not None
        if value is None:
            if feature.typ.is_nullable:
                return None
            else:
                raise ValueError(f"Value is none but feature {feature} is not nullable")
        if feature.typ.collection_type is not None and get_origin(feature.typ.collection_type) == set:
            if not isinstance(value, set):
                raise TypeError(f"Feature {feature} is a set but value {value} is not")
            return {self._default_decode_value(feature, x) for x in value}
        elif feature.typ.collection_type is not None and get_origin(feature.typ.collection_type) == list:
            if not isinstance(value, list):
                raise TypeError(f"Feature {feature} is a list but value {value} is not")
            return [self._default_decode_value(feature, x) for x in value]
        else:
            return self._default_decode_value(feature, value)

    def decode(
        self,
        feature: Feature,
        value: Any,
    ):
        if value is None:
            return None

        if feature.decoder is not None:
            return feature.decoder(value)

        return self._default_decode(feature, value)

    def decode_fqn(
        self,
        fqn: str,
        value: Any,
    ):
        from chalk.features import Feature

        return self.decode(Feature.from_root_fqn(fqn), value)

    def get_pandas_dtype(self, fqn: str) -> str:
        from chalk.features import Feature

        feature = Feature.from_root_fqn(fqn)
        typ = feature.typ
        assert typ is not None, "typ should be specified"
        underlying = typ.underlying
        if issubclass(underlying, enum.Enum):
            # For enums, require all members to have the same type
            underlying = get_enum_value_type(underlying)
        # See https://pandas.pydata.org/docs/user_guide/basics.html#basics-dtypes
        if issubclass(underlying, str):
            return "string"
        if issubclass(underlying, bool):
            return "boolean"
        if issubclass(underlying, int):
            return "Int64"
        if issubclass(underlying, float):
            return "Float64"
        if issubclass(underlying, datetime):
            # This assumes timezone-aware. For timezone-unaware the `pandas_dtype` must be set directly on the Feature
            return "datetime64[ns, utc]"
        _log.info(
            f"Defaulting to pandas type 'object' for fqn {fqn} of type {typ.underlying.__name__}. Set the `pandas_dtype` attribute for better specificity."
        )
        return "object"

    def get_polars_dtype(self, fqn: str) -> Union[pl.DataType, Type[pl.DataType]]:
        from chalk.features import Feature

        feature = Feature.from_root_fqn(fqn)
        if feature.polars_dtype is not None:
            return feature.polars_dtype
        typ = feature.typ
        assert typ is not None, "typ should be specified"
        typ.collection_type
        underlying = typ.underlying
        dtype = self._get_pl_dtype_for_type(underlying, fqn)
        if typ.collection_type is not None:
            dtype = pl.List(dtype)
        return dtype

    def _get_pl_dtype_for_type(self, underlying: type, fqn: str) -> Union[pl.DataType, Type[pl.DataType]]:
        if issubclass(underlying, enum.Enum):
            # For enums, require all members to have the same type
            underlying = get_enum_value_type(underlying)
        if issubclass(underlying, str):
            return pl.Utf8
        if issubclass(underlying, bool):
            return pl.Boolean
        if issubclass(underlying, int):
            return pl.Int64
        if issubclass(underlying, float):
            return pl.Float64
        if issubclass(underlying, datetime):
            return pl.Datetime(time_unit="us", time_zone="UTC")
        if issubclass(underlying, date):
            return pl.Date
        if dataclasses.is_dataclass(underlying) or is_namedtuple(underlying, annotated=True):
            # Polars always treats dataclasses and namedtuples as structs
            annotations = get_type_hints(underlying)
            return pl.Struct(
                fields=[
                    pl.Field(name=k, dtype=self._get_pl_dtype_for_type(v, f"{fqn}.{k}"))
                    for (k, v) in annotations.items()
                ]
            )
        _log.info(
            f"Defaulting to polars type 'object' for fqn {fqn} of type {underlying.__name__}. Set the `polars_dtype` attribute for better specificity."
        )
        return pl.Object

    def get_sqlalchemy_type(self, fqn: str) -> Union[TypeEngine, Type[TypeEngine]]:
        from chalk.features import Feature

        feature = Feature.from_root_fqn(fqn)
        if feature.sqlalchemy_dtype is not None:
            return feature.sqlalchemy_dtype
        typ = feature.typ
        assert typ is not None, "typ should be specified"
        underlying = typ.underlying
        if issubclass(underlying, enum.Enum):
            # For enums, require all members to have the same type
            underlying = get_enum_value_type(underlying)
        # See https://pandas.pydata.org/docs/user_guide/basics.html#basics-dtypes
        if issubclass(underlying, str):
            return Text
        if issubclass(underlying, bool):
            return Boolean
        if issubclass(underlying, int):
            return Integer
        if issubclass(underlying, float):
            return Float
        if issubclass(underlying, datetime):
            # This type requires the local timezone attribute to be set, but it will be stored as
            # a timezone-unaware timestamp in UTC. It will be converted back into a UTC tz-aware object
            # when parsed by python
            return UtcDateTime
        raise TypeError(
            f"Unable to infer a SQLAlchemy Dtype for field {fqn}. Please set the `sqlalchemy_dtype` attribute."
        )


FEATURE_CODEC = FeatureCodec()



import inspect
import logging
from typing import Callable, Iterable, List, Optional, Set, TypeVar, Union, overload

T = TypeVar("T")


HookFn = Callable[[], T]
EnvironmentFilter = Optional[Union[List[str], str]]


def _run_all_hooks(environment: str, hooks: Iterable["Hook"]) -> None:
    for hook in hooks:
        if hook.environment is None or environment in hook.environment:
            try:
                hook()
            except Exception as e:
                logging.error(f"Error running hook {hook.fn.__name__}")
                raise e


class Hook:
    # Registry
    before_all: Set["Hook"] = set()
    after_all: Set["Hook"] = set()

    environment: Optional[List[str]]
    fn: HookFn
    filename: str

    def __init__(self, fn: HookFn, filename: str, environment: Optional[EnvironmentFilter] = None):
        self.fn = fn
        self.filename = filename
        self.environment = (
            [
                environment,
            ]
            if isinstance(environment, str)
            else environment
        )

    def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    def __repr__(self):
        return f'Hook(filename={self.filename}, fn={self.fn.__name__}", environment={str(self.environment)})'

    @classmethod
    def run_all_before_all(cls, environment: str) -> None:
        return _run_all_hooks(environment, cls.before_all)

    @classmethod
    def run_all_after_all(cls, environment: str) -> None:
        return _run_all_hooks(environment, cls.after_all)


@overload
def before_all(fn: HookFn, /) -> Hook:
    ...


@overload
def before_all(fn: None = None, /, environment: EnvironmentFilter = None) -> Callable[[HookFn], Hook]:
    ...


def before_all(
    fn: Optional[HookFn] = None, /, environment: EnvironmentFilter = None
) -> Union[Hook, Callable[[HookFn], Hook]]:
    caller_filename = inspect.stack()[1].filename

    def decorator(f, cf=caller_filename):
        hook = Hook(fn=f, filename=cf, environment=environment)
        Hook.before_all.add(hook)
        return hook

    return decorator(fn) if fn else decorator


@overload
def after_all(fn: HookFn, /, environment: EnvironmentFilter = None) -> Hook:
    ...


@overload
def after_all(fn: None = None, /, environment: EnvironmentFilter = None) -> Callable[[HookFn], Hook]:
    ...


def after_all(
    fn: Optional[HookFn] = None, /, environment: EnvironmentFilter = None
) -> Union[Hook, Callable[[HookFn], Hook]]:
    caller_filename = inspect.stack()[1].filename

    def decorator(f, cf=caller_filename):
        hook = Hook(fn=f, filename=cf, environment=environment)
        Hook.after_all.add(hook)
        return hook

    return decorator(fn) if fn else decorator



import inspect
from typing import Any, Callable, Dict, List, Optional, Union, get_type_hints

from pydantic import BaseModel

from chalk import MachineType
from chalk.features import Features
from chalk.features.resolver import ResolverAnnotationParser
from chalk.streams._types import StreamResolver, StreamSource


def _parse_stream_resolver(
    *,
    caller_globals: Optional[Dict[str, Any]],
    caller_locals: Optional[Dict[str, Any]],
    fn: Callable,
    source: StreamSource,
    caller_filename: str,
    environment: Optional[Union[List[str], str]] = None,
    machine_type: Optional[MachineType] = None,
) -> StreamResolver:
    fqn = f"{fn.__module__}.{fn.__name__}"
    function_definition = inspect.getsource(fn)
    signature = inspect.signature(fn)
    annotation_parser = ResolverAnnotationParser(fn, caller_globals, caller_locals)
    inputs = [annotation_parser.parse_annotation(p) for p in signature.parameters.keys()]
    if len(inputs) != 1 or not issubclass(inputs[0], (BaseModel, str)):
        raise TypeError(f"Stream resolver '{fqn}' must take a single input, either a str or pydantic model")

    input_type = inputs[0]
    return_annotation = get_type_hints(fn)["return"]

    if not isinstance(return_annotation, type):
        raise TypeError(f"return_annotation {return_annotation} of type {type(return_annotation)} is not a type")
    if not issubclass(return_annotation, Features):
        raise ValueError(f"Stream resolver '{fqn}' did not have a valid return type: must be Features[]")

    # function annotated like def get_account_id(user_id: User.id) -> Features[User.account_id]
    # or def get_account_id(user_id: User.id) -> User:
    output_features = return_annotation
    resolver = StreamResolver(
        filename=caller_filename,
        source=source,
        function_definition=function_definition,
        fqn=fqn,
        doc=fn.__doc__,
        fn=fn,
        model=input_type if isinstance(input_type, BaseModel) else None,
        output_features=output_features,
        environment=environment,
        machine_type=machine_type,
    )

    StreamResolver.registry.append(resolver)
    return resolver


def stream(
    *,
    source: StreamSource,
    environment: Optional[Union[List[str], str]] = None,
    machine_type: Optional[MachineType] = None,
):
    caller_frame = inspect.stack()[1]
    caller_filename = caller_frame.filename
    caller_globals = caller_frame.frame.f_globals
    caller_locals = caller_frame.frame.f_locals

    def decorator(fn: Callable):
        return _parse_stream_resolver(
            caller_globals=caller_globals,
            caller_locals=caller_locals,
            fn=fn,
            source=source,
            caller_filename=caller_filename,
            environment=environment,
            machine_type=machine_type,
        )

    return decorator



import math
import pathlib
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from typing import Optional

import pandas as pd
import polars as pl
import polars.testing
import pytest

from chalk.features import DataFrame, Filter, after, feature_time, features, has_many, has_one, unwrap_feature
from chalk.features.resolver import ResolverAnnotationParser


@dataclass(frozen=True)
class Foo:
    bar: int


@features
class ToppingPrice:
    id: str
    topping_id: str
    wow: str
    ts: datetime = feature_time()


@features
class Nickname:
    id: str


@features
class Taco:
    id: str
    user_id: str
    price: int
    maybe_price: Optional[int]
    hat: str
    topping_id: str
    ts: datetime = feature_time()
    nicknames: DataFrame[Nickname] = has_many(lambda: Nickname.id == Taco.id)
    topping: "Topping" = has_one(lambda: Topping.id == Taco.topping_id)
    foo: Foo


@features
class Topping:
    id: str
    value: int
    ts: datetime = feature_time()
    nicknames: DataFrame[Nickname] = has_many(lambda: Nickname.id == Topping.id)
    price: ToppingPrice = has_one(lambda: ToppingPrice.topping_id == Topping.id)


class TestDF:
    def test_indexing(self):
        f = DataFrame({Taco.id: [1, 2, 3], Taco.price: [5, 6, 7]})
        polars.testing.assert_frame_equal(
            f.to_polars().collect()[-1],
            DataFrame(
                {
                    Taco.id: [3],
                    Taco.price: [7],
                }
            )
            .to_polars()
            .collect(),
        )

    def test_range(self):
        f = DataFrame({Taco.id: [1, 2, 3]})
        assert len(f[0:]) == len(f) == 3
        assert len(f[1:]) == 2
        assert len(f[0:1]) == 1

    def test_eq(self):
        assert (DataFrame.from_dict({Taco.id: [1, 1, 3]}) == DataFrame.from_dict({Taco.id: [1, 1, 3]})).all()
        assert not (DataFrame.from_dict({Taco.id: [1, 1, 2]}) == DataFrame.from_dict({Taco.id: [1, 1, 3]})).all()

    def test_from_pandas(self):
        df = pd.DataFrame({Taco.id: ["1", "2", "3"], Taco.price: [5, 6, 7]})
        chalk_df = DataFrame(df)
        is_eq = chalk_df.to_pandas() == df
        is_all_eq = is_eq.all().all()
        assert is_all_eq

    def test_empty(self):
        empty_df = DataFrame()
        assert empty_df.references_feature_set is None

    def test_datetimes_timezone(self):
        now_ts = datetime.now(timezone.utc)
        df = DataFrame(
            {
                Taco.ts: [now_ts],
                Taco.id: [1],
            }
        )
        item = df[Taco.ts].item()
        assert isinstance(item, datetime)
        assert item == now_ts

    def test_in(self):
        chalk_df = DataFrame(
            {
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.price: [1, 2, 10],
            }
        )
        one_two_three_tacos = chalk_df[Taco.price in (1, 2, 3)]
        assert isinstance(one_two_three_tacos, DataFrame)
        assert len(one_two_three_tacos) == 2
        assert one_two_three_tacos[Taco.price].max() == 2
        assert one_two_three_tacos[Taco.price].min() == 1
        in_tacos_1 = chalk_df[Taco.price in []]
        assert len(in_tacos_1) == 0
        in_tacos_2 = chalk_df[Taco.price in [444]]
        assert len(in_tacos_2) == 0
        in_tacos_3 = chalk_df[Taco.price not in [1, 4]]
        assert len(in_tacos_3) == 2
        in_tacos_4 = chalk_df[not Taco.price in [1]]
        assert len(in_tacos_4) == 2

    def test_struct(self):
        df = DataFrame(
            {
                Taco.foo: [Foo(1), Foo(2), Foo(3)],
                Taco.id: [1, 2, 3],
            }
        )
        ans_1 = df[Taco.foo == Foo(1), Taco.id]
        assert (ans_1 == DataFrame({Taco.id: ["1"]})).all()
        ans_2 = df[Taco.foo != Foo(1), Taco.id]
        assert (ans_2 == DataFrame({Taco.id: [2, 3]})).all()
        ans_3 = df[Taco.foo in [Foo(1)], Taco.id]
        assert (ans_3 == DataFrame({Taco.id: [1]})).all()
        ans_4 = df[Taco.foo not in [Foo(1)], Taco.id]
        assert (ans_4 == DataFrame({Taco.id: [2, 3]})).all()

    def test_is_none(self):
        chalk_df = DataFrame(
            {
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.maybe_price: [20, None, 10],
            }
        )
        is_none = chalk_df[Taco.maybe_price is None]
        assert len(is_none) == 1
        equals_none = chalk_df[Taco.maybe_price == None]
        assert len(equals_none) == 1
        is_not_none = chalk_df[Taco.maybe_price is not None]
        assert len(is_not_none) == 2
        ne_none = chalk_df[Taco.maybe_price != None]
        assert len(ne_none) == 2

    def test_bool_op(self):
        chalk_df = DataFrame(
            {
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.maybe_price: [20, None, 10],
                Taco.price: [1, 2, 3],
            }
        )
        tacos = chalk_df[Taco.maybe_price is None or Taco.id == "t_1"]
        assert len(tacos) == 2
        tacos = chalk_df[Taco.maybe_price is None and Taco.id == "t_2"]
        assert len(tacos) == 1
        tacos = chalk_df[not (Taco.maybe_price is not None or not Taco.id == "t_2")]
        assert len(tacos) == 1

        tacos = chalk_df[Taco.maybe_price is None and Taco.id == "t_1"]
        assert len(tacos) == 0
        tacos = chalk_df[Taco.maybe_price is not None and (Taco.id == "t_1" or Taco.price == 3)]
        assert len(tacos) == 2
        tacos = chalk_df[Taco.maybe_price is not None and (Taco.id == "t_1" or Taco.price in [8, 4, 99])]
        assert len(tacos) == 1

    def test_after(self):
        now = datetime.now()
        data = [now - timedelta(days=10), now]
        series = pl.Series(values=data, dtype=pl.Datetime)
        chalk_df = DataFrame(
            {
                Taco.ts: series,
                Taco.price: [5, 10],
            }
        )
        new_price = chalk_df[after(days_ago=5), Taco.price].sum()
        assert new_price == 10

    def test_filtering(self):
        chalk_df = DataFrame(
            {
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.price: [1, 5, 10],
            }
        )
        expensive_tacos = chalk_df[Taco.price > 1]
        assert isinstance(expensive_tacos, DataFrame)
        assert len(expensive_tacos) == 2
        assert expensive_tacos[Taco.price].max() == 10
        assert expensive_tacos[Taco.price].min() == 5

    def test_filter_and_project(self):
        chalk_df = DataFrame(
            {
                Taco.user_id: ["u_1", "u_1", "u_2"],
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.price: [1, 5, 10],
            }
        )
        expensive_tacos = chalk_df[Taco.price, Taco.price > 1, Taco.user_id == "u_1"]
        assert len(expensive_tacos) == 1
        assert expensive_tacos.max() == 5
        assert expensive_tacos.min() == 5

    def test_filter_many(self):
        chalk_df = DataFrame(
            {
                Taco.user_id: ["u_1", "u_1", "u_2"],
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.price: [1, 5, 10],
            }
        )
        expensive_tacos = chalk_df[Taco.price > 1, Taco.user_id == "u_1"]
        assert len(expensive_tacos) == 1
        assert expensive_tacos[Taco.price].max() == 5
        assert expensive_tacos[Taco.price].min() == 5

    def test_filtering_with_datetime(self):
        chalk_df = DataFrame(
            {
                Taco.user_id: ["u_1", "u_1", "u_2"],
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.price: [1, 5, 10],
                Taco.ts: [
                    datetime.now(tz=timezone.utc) - timedelta(days=3),
                    datetime.now(tz=timezone.utc) - timedelta(days=1),
                    datetime.now(tz=timezone.utc),
                ],
            }
        )
        recent_tacos = chalk_df[Taco.ts >= datetime.now(tz=timezone.utc) - timedelta(hours=3)]
        assert isinstance(recent_tacos, DataFrame)
        assert len(recent_tacos) == 1
        assert recent_tacos[Taco.price].max() == 10

    def test_indexing_one(self):
        chalk_df = DataFrame(
            {
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.price: [100, 200, 300],
            }
        )
        assert chalk_df[Taco.price].mean() == 200
        assert len(chalk_df[Taco.price]) == 3
        assert chalk_df[Taco.price].max() == 300
        assert chalk_df[Taco.price].min() == 100

    def test_indexing_many(self):
        chalk_df = DataFrame(
            {
                Taco.id: ["t_1", "t_2", "t_3"],
                Taco.price: [100, 200, 300],
                Taco.hat: ["green", "red", "blue"],
            }
        )
        actual = chalk_df[Taco.price, Taco.id]
        expected = DataFrame(
            {
                Taco.price: [100, 200, 300],
                Taco.id: ["t_1", "t_2", "t_3"],
            }
        )
        polars.testing.assert_frame_equal(actual.to_polars(), expected.to_polars())

    def test_input_feature_types(self):
        df = DataFrame({Taco.topping.id: ["1"]})
        assert df[Taco.topping.id].item() == "1"

    def test_from_dict(self):
        df = DataFrame(
            {
                Taco.topping.value: [100, 200],
            }
        )
        assert 300 == df[Taco.topping.value].sum()

    def test_from_list(self):
        assert 0 == len(DataFrame.from_list([]))
        assert len(DataFrame.from_list(Taco(price=123, user_id="33"))) == 1
        assert len(DataFrame.from_list([Taco(price=123, user_id="33")])) == 1
        assert len(DataFrame.from_list(Taco(price=123, user_id="33"), Taco(price=3, user_id="33"))) == 2
        assert len(DataFrame.from_list([Taco(price=123, user_id="33"), Taco(price=3, user_id="33")])) == 2
        with pytest.raises(ValueError):
            DataFrame.from_list(Taco(price=123, user_id="33"), Taco(price=3))

        with pytest.raises(ValueError):
            DataFrame.from_list(Taco(), Taco(price=3, user_id="33"))

    def test_ast_parsing(self):
        container = {1, 2, 3}

        def my_resolver(
            my_arg: DataFrame[
                Taco.price,
                Taco.user_id > "33",
                Taco.maybe_price is not None,
                Taco.maybe_price is None,
                Taco.price not in container,
                Taco.price in container or (Taco.topping.price == -3 and (Taco.hat == "cool" or Taco.price == 4)),
            ]
        ):
            ...

        fancy_index = ResolverAnnotationParser(my_resolver, globals(), locals()).parse_annotation("my_arg")
        assert issubclass(fancy_index, DataFrame)

        assert ">" == fancy_index.filters[0].operation
        assert "taco.user_id" == str(fancy_index.filters[0].lhs)
        assert "33" == fancy_index.filters[0].rhs

        assert "!=" == fancy_index.filters[1].operation
        assert "taco.maybe_price" == str(fancy_index.filters[1].lhs)
        assert None == fancy_index.filters[1].rhs

        assert "==" == fancy_index.filters[2].operation
        assert "taco.maybe_price" == str(fancy_index.filters[2].lhs)
        assert None == fancy_index.filters[2].rhs

        assert "not in" == fancy_index.filters[3].operation
        assert "taco.price" == str(fancy_index.filters[3].lhs)
        assert container == fancy_index.filters[3].rhs

        fancy_clause = fancy_index.filters[4]
        assert isinstance(fancy_clause, Filter)
        first_fancy = fancy_clause.lhs
        assert isinstance(first_fancy, Filter)
        assert "in" == first_fancy.operation
        assert "taco.price" == str(first_fancy.lhs)
        assert container == first_fancy.rhs

        assert "or" == fancy_clause.operation

    def test_to_pandas(self):
        df = DataFrame({ToppingPrice.id: ["hi"]})
        pandas_df = df.to_pandas()
        assert isinstance(pandas_df, pd.DataFrame)
        assert pandas_df.dtypes.tolist() == [pd.StringDtype()]
        assert pandas_df.columns.tolist() == [unwrap_feature(ToppingPrice.id)]
        assert pandas_df.values.tolist() == [["hi"]]

    def test_wrong_dtype(self):
        with pytest.raises(TypeError, match="Values for feature `topping.value` could not be converted to dtype `i64`"):
            # Should raise if constructing via a dict
            DataFrame({Topping.value: ["not_an_int"]})

        with pytest.raises(TypeError, match="Values for feature `topping.value` could not be converted to dtype `i64`"):
            # Should also raise if passing in another dataframe
            DataFrame(pl.DataFrame({"topping.value": ["not_an_int"]}))

        lf = DataFrame(pl.DataFrame({"topping.value": ["not_an_int"]}).lazy())
        with pytest.raises(
            pl.ComputeError,
        ):
            # But if passing a lazy frame, we won't get the error until we collect
            # And it will be a generic polars ComputeError
            lf.to_polars().collect()

    def test_item(self):
        df = DataFrame({Topping.value: [2]})
        assert df.item() == 2

    def test_float(self):
        df = DataFrame({Topping.value: [2]})
        assert float(df) == 2.0

    def test_int(self):
        df = DataFrame({Topping.value: [2]})
        assert int(df) == 2

    def test_aggregation(self):
        df = DataFrame({Topping.value: [2, 3]})
        assert df.mean() == 2.5
        assert df.sum() == 5
        assert df.min() == 2
        assert df.max() == 3
        assert df.median() == 2.5
        assert df.var() == 0.5
        assert df.std() == math.sqrt(0.5)

    def test_csv(self, tmp_path: pathlib.Path):
        df = DataFrame({Topping.value: [2, 3]})
        pl_df = df.to_polars().collect()
        pl_df = pl_df.rename({"topping.value": "my_col_name"})
        csv_filepath = tmp_path / "my_file.csv"
        pl_df.write_csv(csv_filepath, has_header=True)
        new_df = DataFrame.read_csv(csv_filepath, has_header=True, columns={"my_col_name": "topping.value"})
        assert (df == new_df).all()

    def test_parquet(self, tmp_path: pathlib.Path):
        df = DataFrame({Topping.value: [2, 3]})
        pl_df = df.to_polars().collect()
        pl_df = pl_df.rename({"topping.value": "my_col_name"})
        filepath = tmp_path / "my_file.csv"
        pl_df.write_parquet(filepath)
        new_df = DataFrame.read_parquet(filepath, columns={"my_col_name": "topping.value"})
        assert (df == new_df).all()



from typing import List, Optional, Union

from pydantic import BaseModel

from chalk.streams._types import StreamSource


class KafkaSource(BaseModel, StreamSource):
    broker: Union[str, List[str]]
    topic: Union[str, List[str]]
    ssl_keystore_location: Optional[str] = None
    client_id_prefix: Optional[str] = None
    group_id_prefix: Optional[str] = None
    topic_metadata_refresh_interval_ms: Optional[int] = None
    security_protocol: Optional[str] = None

    def config_to_json(self):
        return self.json(exclude={"message_model"})



import os
from pathlib import Path


def get_directory_root() -> Path:
    root_dir = Path(os.path.dirname(os.path.abspath("dummy.txt")))
    has_init = lambda p: (p / "__init__.py").exists()
    while has_init(root_dir.parent):
        root_dir = root_dir.parent
    return root_dir.parent if has_init(root_dir) else root_dir
